import enum
from typing import Type

from ophyd_async.core import (
    ConfigSignal,
    DeviceVector,
    SignalRW,
    StandardReadable,
    SubsetEnum,
    T,
)
from ophyd_async.epics.signal import epics_signal_r, epics_signal_rw, epics_signal_x


class StrEnum(str, enum.Enum):
    pass


def epics_signal_io(datatype: Type[T], prefix: str, name: str = "") -> SignalRW[T]:
    """Create a `SignalRW` backed by 2 EPICS PVs.

    The write PV gets an extra 'O' and the read PV gets an extra 'I'
    added to the prefix.

    Parameters
    ----------
    datatype:
        Check that the PV is of this type
    prefix:
        The PV to read and monitor

    """
    return epics_signal_rw(
        datatype, read_pv=f"{prefix}I", write_pv=f"{prefix}O", name=name
    )


class DG645Channel(StandardReadable):
    Reference = SubsetEnum["T0", "A", "B", "C", "D", "E", "F", "G", "H"]

    def __init__(self, prefix: str, name: str = ""):
        with self.add_children_as_readables(ConfigSignal):
            self.reference = epics_signal_io(self.Reference, f"{prefix}ReferenceM")
            self.delay = epics_signal_io(float, f"{prefix}DelayA")
        super().__init__(name=name)


class DG645Output(StandardReadable):
    class Polarity(StrEnum):
        NEG = "NEG"
        POS = "POS"

    def __init__(self, prefix: str, name: str = ""):
        with self.add_children_as_readables(ConfigSignal):
            self.polarity = epics_signal_io(self.Polarity, f"{prefix}OutputPolarityB")
            self.amplitude = epics_signal_io(float, f"{prefix}OutputAmpA")
            self.offset = epics_signal_io(float, f"{prefix}OutputOffsetA")
        self.output_mode_ttl = epics_signal_x(f"{prefix}OutputModeTtlSS.PROC")
        self.output_mode_nim = epics_signal_x(f"{prefix}OutputModeNimSS.PROC")
        super().__init__(name=name)


class DG645DelayOutput(DG645Output):
    def __init__(self, prefix: str, name: str = ""):
        with self.add_children_as_readables(ConfigSignal):
            self.trigger_prescale = epics_signal_io(int, f"{prefix}TriggerPrescaleL")
            self.trigger_phase = epics_signal_io(int, f"{prefix}TriggerPhaseL")
        super().__init__(prefix=prefix, name=name)


class DG645Delay(StandardReadable):
    def __init__(self, prefix: str, name: str = ""):
        # Conventional signals
        with self.add_children_as_readables(ConfigSignal):
            self.label = epics_signal_rw(str, f"{prefix}Label")
            self.device_id = epics_signal_r(str, f"{prefix}IdentSI")
        self.status = epics_signal_r(str, f"{prefix}StatusSI")
        self.clear_error = epics_signal_x(f"{prefix}StatusClearBO")
        self.goto_remote = epics_signal_x(f"{prefix}GotoRemoteBO")
        self.goto_local = epics_signal_x(f"{prefix}GotoLocalBO")
        self.reset = epics_signal_x(f"{prefix}ResetBO")
        self.status_checking = epics_signal_rw(bool, f"{prefix}StatusCheckingBO")
        self.reset_serial = epics_signal_x(f"{prefix}IfaceSerialResetBO")
        self.serial_state = epics_signal_io(bool, f"{prefix}IfaceSerialStateB")
        self.serial_baud = epics_signal_io(
            SubsetEnum["4800", "9600", "19200", "38400", "57600", "115200"],
            f"{prefix}IfaceSerialBaudM",
        )
        self.reset_gpib = epics_signal_x(f"{prefix}IfaceGpibResetBO")
        self.gpib_state = epics_signal_io(bool, f"{prefix}IfaceGpibStateB")
        self.gpib_address = epics_signal_io(int, f"{prefix}IfaceGpibAddrL")
        self.reset_lan = epics_signal_x(f"{prefix}IfaceLanResetBO")
        self.mac_address = epics_signal_r(str, f"{prefix}IfaceMacAddrSI")
        self.lan_state = epics_signal_io(bool, f"{prefix}IfaceLanStateB")
        self.dhcp_state = epics_signal_io(bool, f"{prefix}IfaceDhcpStateB")
        self.autoip_state = epics_signal_io(bool, f"{prefix}IfaceAutoIpStateB")
        self.static_ip_state = epics_signal_io(bool, f"{prefix}IfaceStaticIpStateB")
        self.bare_socket_state = epics_signal_io(bool, f"{prefix}IfaceBareSocketStateB")
        self.telnet_state = epics_signal_io(bool, f"{prefix}IfaceTelnetStateB")
        self.vxi11_state = epics_signal_io(bool, f"{prefix}IfaceVxiStateB")
        self.ip_address = epics_signal_io(str, f"{prefix}IfaceIpAddrS")
        self.network_mask = epics_signal_io(str, f"{prefix}IfaceNetMaskS")
        self.gateway = epics_signal_io(str, f"{prefix}IfaceGatewayS")
        # Individual delay channels
        with self.add_children_as_readables():
            self.channels = DeviceVector(
                {
                    "A": DG645Channel(f"{prefix}A"),
                    "B": DG645Channel(f"{prefix}B"),
                    "C": DG645Channel(f"{prefix}C"),
                    "D": DG645Channel(f"{prefix}D"),
                    "E": DG645Channel(f"{prefix}E"),
                    "F": DG645Channel(f"{prefix}F"),
                    "G": DG645Channel(f"{prefix}G"),
                    "H": DG645Channel(f"{prefix}H"),
                }
            )
        # 2-channel delay outputs
        with self.add_children_as_readables():
            self.outputs = DeviceVector(
                {
                    "T0": DG645Output(f"{prefix}T0"),
                    "AB": DG645DelayOutput(f"{prefix}AB"),
                    "CD": DG645DelayOutput(f"{prefix}CD"),
                    "EF": DG645DelayOutput(f"{prefix}EF"),
                    "GH": DG645DelayOutput(f"{prefix}GH"),
                }
            )
        # Trigger control
        with self.add_children_as_readables(ConfigSignal):
            self.trigger_source = epics_signal_io(
                SubsetEnum[
                    "Internal",
                    "Ext rising edge",
                    "Ext falling edge",
                    "SS ext rise edge",
                    "SS ext fall edge",
                    "Single shot",
                    "Line",
                ],
                f"{prefix}TriggerSourceM",
            )
            self.trigger_inhibit = epics_signal_io(
                SubsetEnum["Off", "Triggers", "AB", "AB,CD", "AB,CD,EF", "AB,CD,EF,GH"],
                f"{prefix}TriggerInhibitM",
            )
            self.trigger_level = epics_signal_io(float, f"{prefix}TriggerLevelA")
            self.trigger_rate = epics_signal_io(float, f"{prefix}TriggerRateA")
            self.trigger_advanced_mode = epics_signal_io(
                bool, f"{prefix}TriggerAdvancedModeB"
            )
            self.trigger_holdoff = epics_signal_io(float, f"{prefix}TriggerHoldoffA")
            self.trigger_prescale = epics_signal_io(int, f"{prefix}TriggerPrescaleL")
        # Burst settings
        with self.add_children_as_readables(ConfigSignal):
            self.burst_mode = epics_signal_io(bool, f"{prefix}BurstModeB")
            self.burst_count = epics_signal_io(int, f"{prefix}BurstCountL")
            self.burst_config = epics_signal_io(
                SubsetEnum["All Cycles", "1st Cycle"], f"{prefix}BurstConfigB"
            )
            self.burst_delay = epics_signal_io(float, f"{prefix}BurstDelayA")
            self.burst_period = epics_signal_io(float, f"{prefix}BurstPeriodA")
        super().__init__(name=name)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2024, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import pytest

from firefly.tweak import TweakDisplay


@pytest.fixture()
def display(qtbot):
    disp = TweakDisplay()
    qtbot.addWidget(disp)
    return disp


def test_value_updates_button_steps(display):
    display.ui.value_spin_box.setValue(6.28)
    assert display.ui.reverse_button.pressValue == "-6.28"
    assert display.ui.forward_button.pressValue == "6.28"



import logging

import databroker
import IPython
from bluesky import RunEngine as BlueskyRunEngine
from bluesky.callbacks.best_effort import BestEffortCallback
from bluesky.utils import ProgressBarManager, register_transform

from .exceptions import ComponentNotFound
from .instrument import beamline
from .preprocessors import inject_haven_md_wrapper

log = logging.getLogger(__name__)


catalog = None


def save_data(name, doc):
    # This is a hack around a problem with garbage collection
    # Has been fixed in main, maybe released in databroker v2?
    # Create the databroker callback if necessary
    global catalog
    if catalog is None:
        catalog = databroker.catalog["bluesky"]
    # Save the document
    catalog.v1.insert(name, doc)


def run_engine(connect_databroker=True, use_bec=True, **kwargs) -> BlueskyRunEngine:
    RE = BlueskyRunEngine(**kwargs)
    # Add the best-effort callback
    if use_bec:
        RE.subscribe(BestEffortCallback())
    # Install suspenders
    try:
        aps = beamline.registry.find("APS")
    except ComponentNotFound:
        log.warning("APS device not found, suspenders not installed.")
    else:
        # Suspend when shutter permit is disabled
        # Re-enable when the APS shutter permit signal is better understood
        pass
        # RE.install_suspender(
        #     suspenders.SuspendWhenChanged(
        #         signal=aps.shutter_permit,
        #         expected_value="PERMIT",
        #         allow_resume=True,
        #         sleep=3,
        #         tripped_message="Shutter permit revoked.",
        #     )
        # )
    # Add a shortcut for using the run engine more efficiently
    RE.waiting_hook = ProgressBarManager()
    if (ip := IPython.get_ipython()) is not None:
        register_transform("RE", prefix="<", ip=ip)
    # Install databroker connection
    if connect_databroker:
        RE.subscribe(save_data)
    # Add preprocessors
    RE.preprocessors.append(inject_haven_md_wrapper)
    return RE


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from collections import namedtuple
from typing import Mapping

import matplotlib.pyplot as plt
import numpy as np
from bluesky import plans as bp
from lmfit.models import GaussianModel, StepModel

from ..instrument import beamline


def knife_scan(
    knife_motor,
    start: float,
    end: float,
    num: int,
    I0="I0",
    It="It",
    relative: bool = False,
    md: Mapping = {},
):
    """Plan to scan over a knife placed in the beam to measure its height.

    Parameters
    ==========
    knife_motor
      An ophyd device or name of a registered device that has a well
      defined edge attached to it.
    start
      Motor position at which to start scanning.
    end
      Motor position at which to finish scanning.
    num
      How many points to measure between *start* and *end*.
    I0
      Device or name of a registered ion chamber upstream from the
      knife.
    It
      Device or name of a registered ion chamber downstream from the
      knife.
    relative
      If true, *start* and *end* will be interpreted as relative to
      the current motor position.
    md
      Extra metadata to pass into the run engine.

    """
    md_ = dict(plan_name="knife_scan", **md)
    I0 = beamline.registry.find(name=I0)
    It = beamline.registry.find(name=It)
    knife_motor = beamline.registry.find(name=knife_motor)
    # Check for relative or absolute scan
    if relative:
        plan_func = bp.rel_scan
    else:
        plan_func = bp.scan
    # Compute the plan
    yield from plan_func([I0, It], knife_motor, start, end, num=num, md=md_)


def fit_step(x, y, plot=False, plot_derivative=False):
    """Extract beam properties from a step scan over the beam profile.

    Parameters
    ==========
    x
      Array of motor positions.
    y
      Array of measured values at each motor position
      (e.g. transmitted intensity).
    plot
      If true, plot the fitting results.
    plot_derivative
      If true, also fit and plot the derivative of the step scan.

    Returns
    =======
    properties
      A named tuple with the *size* and *position* of the beam, in
      units of *x*.

    """
    x = np.asarray(x)
    y = np.asarray(y)
    # Normalize data to start at zero
    y -= y[0]
    # Set up initial fitting parameters
    model = StepModel(form="erf")
    params = model.guess(y, x)
    half_max = (np.min(y) + np.max(y)) / 2
    center_idx = np.argmin(np.abs(y - half_max))
    center_guess = x[center_idx]
    params["center"].set(center_guess)
    # Invert the amplitude if we started with full transmission
    is_inverted = np.mean(y) < 0
    if is_inverted:
        params["amplitude"].set(-params["amplitude"].value)
    # Fit the model to the data
    result = model.fit(y, params=params, x=x)
    beam_position = result.values["center"]
    beam_hwhm = result.values["sigma"]
    # Do a fit of the derivative just for testing
    # Plot results
    if plot:
        # Plot knife scan and fit
        fig = plt.figure()
        result.plot(
            fig=fig,
            show_init=False,
            datafmt="x",
            xlabel="Knife position /µm",
            ylabel="Relative transmission",
        )
        ax_res, ax = fig.axes
        line_kw = dict(color="C0", ls=":")
        ax.axvline(beam_position, label="Beam position", **line_kw)
        ax.axhline(half_max, **line_kw)
        ax.text(beam_position, half_max, s=f"{beam_position:.2f}", color="C0")
        ax.axvline(beam_position - beam_hwhm, label="FWHM (y)", **line_kw)
        ax.axvline(beam_position + beam_hwhm, **line_kw)
        ax.legend()
    if plot and plot_derivative:
        # Fit derivative
        model = GaussianModel()
        dy = np.gradient(y, x)
        params = model.guess(dy, x)
        dresult = model.fit(dy, params=params, x=x)
        # Plot derivative of knife scan
        color2 = "C3"
        ax2 = ax.twinx()
        ax2.plot(x, dy, color=color2, label="Derivative", marker="x", ls="None")
        ax2.plot(x, dresult.eval(x=x), color=color2, alpha=0.5, label="Best fit")
        ax2.plot(
            x,
            np.gradient(result.eval(x=x), x),
            color=color2,
            alpha=0.5,
            ls="--",
            label="Step fit gradient",
        )
        dy_half_max = (np.min(dy) + np.max(dy)) / 2
        line_kw["color"] = color2
        ax2.axhline(dy_half_max, **line_kw)
        ax2.axvline(dresult.values["center"], **line_kw)
        dfwhm = dresult.values["fwhm"]
        ax2.axvline(
            dresult.values["center"] - dfwhm / 2, label="FWHM (dy/dx)", **line_kw
        )
        ax2.axvline(dresult.values["center"] + dfwhm / 2, **line_kw)
        ax2.legend()
        ax2.set_ylabel("Derivative /µm⁻")
    # Prepare results object
    Properties = namedtuple("Properties", ["position", "fwhm"])
    properties = Properties(position=beam_position, fwhm=2 * beam_hwhm)
    return properties


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from unittest import mock

from bluesky_queueserver_api import BPlan
from qtpy import QtCore

from firefly.plans.count import CountDisplay


def test_count_plan_queued(qtbot, sim_registry):
    display = CountDisplay()
    display.ui.run_button.setEnabled(True)
    display.ui.num_spinbox.setValue(5)
    display.ui.delay_spinbox.setValue(0.5)
    display.ui.detectors_list.selected_detectors = mock.MagicMock(
        return_value=["vortex_me4", "I0"]
    )
    expected_item = BPlan("count", num=5, detectors=["vortex_me4", "I0"], delay=0.5)

    def check_item(item):
        # from pprint import pprint

        # pprint(item.to_dict())
        # pprint(expected_item.to_dict())
        return item.to_dict() == expected_item.to_dict()

    # Click the run button and see if the plan is queued
    with qtbot.waitSignal(
        display.queue_item_submitted, timeout=1000, check_params_cb=check_item
    ):
        qtbot.mouseClick(display.ui.run_button, QtCore.Qt.LeftButton)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging
from enum import Enum, IntEnum

from ophyd_async.core import (
    ConfigSignal,
    HintedSignal,
    Signal,
    StandardReadable,
    soft_signal_rw,
)
from ophyd_async.epics.signal import epics_signal_r, epics_signal_rw, epics_signal_x

from ..positioner import Positioner
from .signal import derived_signal_r, derived_signal_x

log = logging.getLogger(__name__)


class DoneStatus(IntEnum):
    MOVING = 0
    DONE = 1


class BusyStatus(IntEnum):
    DONE = 0
    BUSY = 1


class MotorDriveStatus(IntEnum):
    NOT_READY = 0
    READY_TO_MOVE = 1


class UndulatorPositioner(Positioner):
    done_value: int = BusyStatus.DONE

    def __init__(
        self,
        prefix: str,
        actuate_signal: Signal,
        stop_signal: Signal,
        done_signal: Signal,
        name: str = "",
    ):
        with self.add_children_as_readables(HintedSignal):
            self.readback = epics_signal_rw(float, f"{prefix}M.VAL")
        with self.add_children_as_readables():
            self.setpoint = epics_signal_rw(float, f"{prefix}SetC.VAL")
        with self.add_children_as_readables(ConfigSignal):
            self.units = epics_signal_r(str, f"{prefix}SetC.EGU")
            self.precision = epics_signal_r(int, f"{prefix}SetC.PREC")
        self.velocity = soft_signal_rw(
            float, initial_value=1
        )  # Need to figure out what this value is
        # Add control signals that depend on the parent
        self.actuate = derived_signal_x(derived_from={"parent_signal": actuate_signal})
        self.stop_signal = derived_signal_x(derived_from={"parent_signal": stop_signal})
        self.done = derived_signal_r(int, derived_from={"parent_signal": done_signal})
        super().__init__(name=name)


class PlanarUndulator(StandardReadable):
    """APS Planar Undulator

    .. index:: Ophyd Device; PlanarUndulator

    The signals *busy* and *done* convey complementary
    information. *busy* comes from the IOC, while *done* comes
    directly from the controller.

    EXAMPLE::

        undulator = PlanarUndulator("S25ID:USID:", name="undulator")

    """

    _ophyd_labels_ = {"xray_sources", "undulators"}

    class AccessMode(str, Enum):
        USER = "User"
        OPERATOR = "Operator"
        MACHINE_PHYSICS = "Machine Physics"
        SYSTEM_MANAGER = "System Manager"

    def __init__(self, prefix: str, name: str = ""):
        # Signals for moving the undulator
        self.start_button = epics_signal_x(f"{prefix}StartC.VAL")
        self.stop_button = epics_signal_x(f"{prefix}StopC.VAL")
        self.busy = epics_signal_r(bool, f"{prefix}BusyM.VAL")
        self.done = epics_signal_r(bool, f"{prefix}BusyDeviceM.VAL")
        self.motor_drive_status = epics_signal_r(int, f"{prefix}MotorDriveStatusM.VAL")
        # Configuration state for the undulator
        with self.add_children_as_readables(ConfigSignal):
            self.harmonic_value = epics_signal_rw(int, f"{prefix}HarmonicValueC")
            self.total_power = epics_signal_r(float, f"{prefix}TotalPowerM.VAL")
            self.gap_deadband = epics_signal_rw(int, f"{prefix}DeadbandGapC")
            self.device_limit = epics_signal_rw(float, f"{prefix}DeviceLimitM.VAL")
            self.device = epics_signal_r(str, f"{prefix}DeviceM")
            self.magnet = epics_signal_r(str, f"{prefix}DeviceMagnetM")
            self.location = epics_signal_r(str, f"{prefix}LocationM")
            self.version_plc = epics_signal_r(float, f"{prefix}PLCVersionM.VAL")
            self.version_hpmu = epics_signal_r(str, f"{prefix}HPMUVersionM.VAL")
        # X-ray spectrum positioners
        with self.add_children_as_readables():
            self.energy = UndulatorPositioner(
                f"{prefix}Energy",
                actuate_signal=self.start_button,
                stop_signal=self.stop_button,
                done_signal=self.busy,
            )
            self.energy_taper = UndulatorPositioner(
                f"{prefix}TaperEnergy",
                actuate_signal=self.start_button,
                stop_signal=self.stop_button,
                done_signal=self.done,
            )
            self.gap = UndulatorPositioner(
                f"{prefix}Gap",
                actuate_signal=self.start_button,
                stop_signal=self.stop_button,
                done_signal=self.done,
            )
            self.gap_taper = UndulatorPositioner(
                f"{prefix}TaperGap",
                actuate_signal=self.start_button,
                stop_signal=self.stop_button,
                done_signal=self.done,
            )
        # Miscellaneous control signals
        self.access_mode = epics_signal_r(self.AccessMode, f"{prefix}AccessSecurityC")
        self.message1 = epics_signal_r(str, f"{prefix}Message1M.VAL")
        self.message2 = epics_signal_r(str, f"{prefix}Message2M.VAL")

        super().__init__(name=name)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from firefly.queue_button import Colors, QueueButton


def test_queue_button_style(qtbot):
    """Does the queue button change color/icon based on the queue state."""
    btn = QueueButton()
    qtbot.addWidget(btn)
    # Initial style should be disabled and plain
    assert not btn.isEnabled()
    assert btn.styleSheet() == ""
    # State when queue server is open and idle (no autostart)
    queue_state = {
        "worker_environment_exists": True,
        "items_in_queue": 0,
        "re_state": "idle",
        "queue_autostart_enabled": False,
    }
    btn.update_queue_style(queue_state)
    assert btn.isEnabled()
    assert Colors.ADD_TO_QUEUE in btn.styleSheet()
    assert btn.text() == "Add to Queue"
    # State when queue server is open and idle (w/ autostart)
    queue_state = {
        "worker_environment_exists": True,
        "items_in_queue": 0,
        "re_state": "idle",
        "queue_autostart_enabled": True,
    }
    btn.update_queue_style(queue_state)
    assert btn.isEnabled()
    assert Colors.RUN_QUEUE in btn.styleSheet()
    assert btn.text() == "Run"
    # State when queue server is open and idle
    queue_state = {
        "worker_environment_exists": True,
        "items_in_queue": 0,
        "re_state": "running",
        "queue_autostart_enabled": True,
    }
    btn.update_queue_style(queue_state)
    assert btn.isEnabled()
    assert Colors.ADD_TO_QUEUE in btn.styleSheet()
    assert btn.text() == "Add to Queue"


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import uuid
from typing import Sequence

from bluesky import Msg
from bluesky import plan_stubs as bps
from ophyd import Device

from ..devices.shutter import ShutterState
from ..instrument import beamline
from .shutters import close_shutters, open_shutters


def count_is_complete(*, old_value, value, **kwargs):
    """Check if the value is done."""
    was_running = old_value == 1
    is_running_now = value == 1
    is_done = was_running and not is_running_now
    print(is_done)
    return is_done


def record_dark_current(
    ion_chambers: Sequence[Device], shutters: Sequence[Device] = []
):
    """Record the dark current on the ion chambers.

    - Close shutters
    - Record ion chamber dark current
    - Restore shutters to their previous positions

    Parameters
    ==========
    ion_chambers
      Ion chamber devices or names.
    shutters
      Shutter devices or names.

    """
    # Get previous shutter states
    old_shutters = {}
    for shutter in shutters:
        old_shutters[shutter] = yield from bps.rd(shutter.readback)
    # Close shutters
    yield from close_shutters(shutters)
    # Measure the dark current
    ion_chambers = beamline.registry.findall(ion_chambers)
    # Record dark currents
    group = uuid.uuid4()
    for ic in ion_chambers:
        yield Msg("trigger", ic, group=group, record_dark_current=True)
        # yield from bps.trigger(ic.record_dark_current, group=group, wait=False)
    # Wait for the devices to be done recording dark current
    yield from bps.wait(group=group)
    # Reset shutters to their original states
    to_open = [
        sht for sht, old_state in old_shutters.items() if old_state == ShutterState.OPEN
    ]
    yield from open_shutters(to_open)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



"""Tests for all the fluorescence detectors.

These tests are mostly parameterized to ensure that both DXP and
Xspress detectors share a common interface. A few of the tests are
specific to one device or another.

"""

import logging
import time
from pathlib import Path

import numpy as np
import pytest
from ophyd import OphydObject, Signal

from haven.devices.dxp import parse_xmap_buffer

DETECTORS = ["dxp", "xspress"]
# DETECTORS = ['dxp']


@pytest.fixture()
def vortex(request):
    """Parameterized fixture for creating a Vortex device with difference
    electronics support.

    """
    # Figure out which detector we're using
    det = request.getfixturevalue(request.param)
    yield det


# @pytest.mark.parametrize("vortex", ["xspress"], indirect=True)
def test_acquire_frames_xspress(xspress):
    """Can we acquire a single frame using the dedicated signal."""
    vortex = xspress
    # Acquire a single frame
    assert vortex.acquire.get() == 0
    vortex.acquire_single.set(1).wait(timeout=3)
    # Check that the num of frames and the acquire button were set
    assert vortex.acquire.get() == 1
    assert vortex.cam.num_images.get() == 1
    assert vortex.acquire_single.get() == 1
    # Does it stop as well
    vortex.acquire_single.set(0).wait(timeout=3)
    assert vortex.acquire.get() == 0
    # Acquire multiple frames
    vortex.acquire_multiple.set(1).wait(timeout=3)
    # Check that the num of frames and the acquire button were set
    assert vortex.acquire.get() == 1
    assert vortex.cam.num_images.get() == 2000
    assert vortex.acquire_single.get() == 1


@pytest.mark.parametrize("vortex", DETECTORS, indirect=True)
def test_roi_size(vortex, caplog):
    """Do the signals for max/size auto-update."""
    roi = vortex.mcas.mca0.rois.roi0
    # Check that we can set the lo_chan without error in the callback
    with caplog.at_level(logging.ERROR):
        roi.lo_chan.set(10).wait(timeout=3)
    for record in caplog.records:
        assert (
            "Another set() call is still in progress" not in record.exc_text
        ), record.exc_text
    # Update the size and check the maximum
    roi.size.set(7).wait(timeout=3)
    assert roi.hi_chan.get() == 17
    # Update the maximum and check the size
    roi.hi_chan.set(28).wait(timeout=3)
    assert roi.size.get() == 18
    # Update the minimum and check the size
    roi.lo_chan.set(25).wait(timeout=3)
    assert roi.size.get() == 3


@pytest.mark.parametrize("vortex", DETECTORS, indirect=True)
def test_roi_size_concurrency(vortex, caplog):
    roi = vortex.mcas.mca0.rois.roi0
    # Set up the roi limits
    roi.lo_chan.set(12).wait(timeout=3)
    roi.size.set(13).wait(timeout=3)
    assert roi.hi_chan.get() == 25
    # Change two signals together
    statuses = [
        roi.lo_chan.set(3),
        roi.hi_chan.set(5),
    ]
    for st in statuses:
        st.wait(timeout=3)
    # Check that the signals were set correctly
    assert roi.lo_chan.get() == 3
    assert roi.hi_chan.get() == 5
    assert roi.size.get() == 2


@pytest.mark.parametrize("vortex", DETECTORS, indirect=True)
def test_enable_some_rois(vortex):
    """Test that the correct ROIs are enabled/disabled."""
    statuses = vortex.enable_rois(rois=[2, 5], elements=[1, 3])
    # Give the IOC time to change the PVs
    for status in statuses:
        status.wait(timeout=3)
        # Check that at least one of the ROIs was changed
    roi = vortex.mcas.mca1.rois.roi2
    is_used = roi.use.get(use_monitor=False)
    assert is_used == 1


@pytest.mark.parametrize("vortex", DETECTORS, indirect=True)
def test_enable_rois(vortex):
    """Test that the correct ROIs are enabled/disabled."""
    statuses = vortex.enable_rois()
    # Give the IOC time to change the PVs
    for status in statuses:
        status.wait(timeout=3)
        # Check that at least one of the ROIs was changed
    roi = vortex.mcas.mca1.rois.roi2
    hinted = roi.use.get(use_monitor=False)
    assert hinted == 1


@pytest.mark.parametrize("vortex", DETECTORS, indirect=True)
def test_disable_some_rois(vortex):
    """Test that the correct ROIs are enabled/disabled."""
    statuses = vortex.enable_rois(rois=[2, 5], elements=[1, 3])
    # Give the IOC time to change the PVs
    for status in statuses:
        status.wait(timeout=3)
    # Check that at least one of the ROIs was changed
    roi = vortex.mcas.mca1.rois.roi2
    hinted = roi.use.get(use_monitor=False)
    assert hinted == 1
    statuses = vortex.disable_rois(rois=[2, 5], elements=[1, 3])
    # Give the IOC time to change the PVs
    for status in statuses:
        status.wait(timeout=3)
    # Check that at least one of the ROIs was changed
    roi = vortex.mcas.mca1.rois.roi2
    hinted = roi.use.get(use_monitor=False)
    assert hinted == 0


@pytest.mark.parametrize("vortex", DETECTORS, indirect=True)
def test_disable_rois(vortex):
    """Test that the correct ROIs are enabled/disabled."""
    statuses = vortex.enable_rois()
    # Give the IOC time to change the PVs
    for status in statuses:
        status.wait(timeout=3)

    statuses = vortex.disable_rois()
    # Give the IOC time to change the PVs
    for status in statuses:
        status.wait(timeout=3)
        # Check that at least one of the ROIs was changed
    roi = vortex.mcas.mca1.rois.roi2
    hinted = roi.use.get(use_monitor=False)
    assert hinted == 0


@pytest.mark.parametrize("vortex", DETECTORS, indirect=True)
def test_stage_signal_names(vortex):
    """Check that we can set the name of the detector ROIs dynamically."""
    dev = vortex.mcas.mca1.rois.roi1
    dev.label.put("~Ni-Ka", timeout=3)
    # Ensure the name isn't changed yet
    assert "Ni-Ka" not in dev.name
    assert "Ni_Ka" not in dev.name
    orig_name = dev.name
    dev.stage()
    try:
        result = dev.read()
    except Exception:
        raise
    else:
        assert "Ni-Ka" not in dev.name  # Make sure it gets sanitized
        assert "~" not in dev.name  # Make sure it gets sanitized
        assert "__" not in dev.name  # Tildes sanitize bad but used for `use` signal
        assert "Ni_Ka" in dev.name
    finally:
        dev.unstage()
    # Name gets reset when unstaged
    assert dev.name == orig_name
    assert dev.count.name == f"{orig_name}_count"
    # Check acquired data uses dynamic names
    for res in result.keys():
        assert "Ni_Ka" in res


@pytest.mark.parametrize("vortex", DETECTORS, indirect=True)
def test_read_and_config_attrs(vortex):
    vortex.mcas.mca0.read_attrs
    expected_read_attrs = [
        "mcas",
        "roi_sums",
        "dead_time_average",
        "dead_time_min",
        "dead_time_max",
    ]
    if hasattr(vortex, "cam"):
        expected_read_attrs.append("cam")
    # Add attrs for summing ROIs across elements
    for roi in range(vortex.num_rois):
        expected_read_attrs.extend(
            [
                f"roi_sums.roi{roi}",
                f"roi_sums.roi{roi}.count",
                f"roi_sums.roi{roi}.net_count",
            ]
        )
    # Add attrs for each MCA and ROI.
    for mca in range(vortex.num_elements):
        expected_read_attrs.extend(
            [
                f"mcas.mca{mca}",
                f"mcas.mca{mca}.rois",
                f"mcas.mca{mca}.spectrum",
                f"mcas.mca{mca}.total_count",
                # f"mcas.mca{mca}.input_count_rate",
                # f"mcas.mca{mca}.output_count_rate",
                f"mcas.mca{mca}.dead_time_percent",
                f"mcas.mca{mca}.dead_time_factor",
                # f"mcas.mca{mca}.background",
            ]
        )
        if hasattr(vortex.mcas.mca0, "clock_ticks"):
            expected_read_attrs.append(f"mcas.mca{mca}.clock_ticks")
        for roi in range(vortex.num_rois):
            expected_read_attrs.extend(
                [
                    f"mcas.mca{mca}.rois.roi{roi}",
                    f"mcas.mca{mca}.rois.roi{roi}.count",
                    f"mcas.mca{mca}.rois.roi{roi}.net_count",
                ]
            )
    assert sorted(vortex.read_attrs) == sorted(expected_read_attrs)


@pytest.mark.parametrize("vortex", DETECTORS, indirect=True)
def test_use_signal(vortex):
    """Check that the ``.use`` ROI signal properly mangles the label.

    It uses label mangling instead of any underlying PVs because
    different detector types don't have this feature or use it in an
    undesirable way.

    """
    roi = vortex.mcas.mca0.rois.roi1
    roi.label.sim_put("Fe-55")
    # Enable the ROI and see if the name is updated
    roi.use.set(False).wait(timeout=3)
    assert roi.label.get() == "~Fe-55"
    # Disable the ROI and see if it goes back
    roi.use.set(True).wait(timeout=3)
    assert roi.label.get() == "Fe-55"
    # Set the label manually and see if the use signal changes
    roi.label.set("~Fe-55").wait(timeout=3)
    assert not bool(roi.use.get())


@pytest.mark.parametrize("vortex", DETECTORS, indirect=True)
def test_stage_hints(vortex):
    """Check that enabled ROIs get hinted."""
    roi0 = vortex.mcas.mca0.rois.roi0
    roi0.label.put("", timeout=3)
    roi0.use.put(1, timeout=3)
    roi1 = vortex.mcas.mca0.rois.roi1
    roi1.label.put("", timeout=3)
    roi1.use.put(0, timeout=3)
    # Ensure the hints aren't applied yet
    assert roi0.count.name not in vortex.hints["fields"]
    assert roi1.count.name not in vortex.hints["fields"]
    # Stage the detector ROIs
    try:
        roi0.stage()
        roi1.stage()
    except Exception:
        raise
    else:
        # Check that only the enabled ROI gets hinted
        assert roi0.count.name in vortex.hints["fields"]
        assert roi1.count.name not in vortex.hints["fields"]
    finally:
        roi0.unstage()
        roi1.unstage()
    # Name gets reset when unstaged
    assert roi0.count.name not in vortex.hints["fields"]
    assert roi1.count.name not in vortex.hints["fields"]


@pytest.mark.skip(reason="DXP fly-scanning not yet implemented")
def test_kickoff_dxp(dxp):
    vortex = dxp
    vortex.write_path = "M:\\tmp\\"
    vortex.read_path = "/net/s20data/sector20/tmp/"
    [
        s.wait(timeout=3)
        for s in [
            vortex.acquiring.set(0),
            vortex.collect_mode.set("MCA Spectrum"),
            vortex.erase_start.set(0),
            vortex.pixel_advance_mode.set("Sync"),
        ]
    ]
    # Ensure that the vortex is in its normal operating state
    assert vortex.collect_mode.get(use_monitor=False) == "MCA Spectrum"
    # Check that the kickoff status ended properly
    status = vortex.kickoff()
    assert not status.done
    vortex.acquiring.set(1)
    status.wait(timeout=3)
    assert status.done
    assert status.success
    # Check that the right signals were set during  kick off
    assert vortex.collect_mode.get(use_monitor=False) == "MCA Mapping"
    assert vortex.erase_start.get(use_monitor=False) == 1
    assert vortex.pixel_advance_mode.get(use_monitor=False) == "Gate"
    # Check that the netCDF writer was setup properly
    assert vortex.net_cdf.enable.get(use_monitor=False) == "Enable"
    assert vortex.net_cdf.file_path.get(use_monitor=False) == "M:\\tmp\\"
    assert vortex.net_cdf.file_name.get(use_monitor=False) == "fly_scan_temp.nc"
    assert vortex.net_cdf.capture.get(use_monitor=False) == 1


def test_dxp_acquire(dxp):
    """Check that the DXP acquire mimics that of the area detector base."""
    assert dxp.stop_all.get(use_monitor=False) == 0
    assert dxp.erase_start.get(use_monitor=False) == 0
    dxp.acquire.set(1).wait(timeout=3)
    assert dxp.stop_all.get(use_monitor=False) == 0
    assert dxp.erase_start.get(use_monitor=False) == 1
    dxp.acquire.set(0).wait(timeout=3)
    assert dxp.stop_all.get(use_monitor=False) == 1
    assert dxp.erase_start.get(use_monitor=False) == 1

    # Now test the reverse behavior
    dxp.acquire.set(0).wait(timeout=3)
    assert dxp.acquire.get(use_monitor=False) == 0
    dxp.acquiring.set(1).wait(timeout=3)
    assert dxp.acquire.get(use_monitor=False) == 1
    dxp.acquiring.set(0).wait(timeout=3)
    assert dxp.acquire.get(use_monitor=False) == 0


@pytest.mark.skip(reason="DXP fly-scanning not yet implemented")
def test_complete_dxp(dxp):
    """Check the behavior of the DXP electornic's fly-scan complete call."""
    vortex = dxp
    vortex.write_path = "M:\\tmp\\"
    vortex.read_path = "/net/s20data/sector20/tmp/"
    vortex.acquire._readback = 1
    status = vortex.complete()
    time.sleep(0.01)
    assert vortex.stop_all.get(use_monitor=False) == 1
    assert not status.done
    vortex.acquiring.set(0)
    status.wait(timeout=3)
    assert status.done


def test_kickoff_xspress(xspress):
    """Check the behavior of the Xspress3 electornic's fly-scan complete call."""
    vortex = xspress
    # Make sure the num_images is included
    fly_sigs = [walk.item for walk in vortex.walk_fly_signals()]
    assert vortex.cam.array_counter in fly_sigs
    # Do the kickoff
    vortex.acquire.sim_put(0)
    status = vortex.kickoff()
    assert not status.done
    # Set the acquire signal to true to test that signals got set
    vortex.detector_state.sim_put(vortex.detector_states.ACQUIRE)
    status.wait(timeout=3)
    assert status.done
    assert vortex.cam.trigger_mode.get() == vortex.trigger_modes.TTL_VETO_ONLY
    assert vortex.acquire.get() == vortex.acquire_states.ACQUIRE


def test_complete_xspress(xspress):
    """Check the behavior of the Xspress3 electornic's fly-scan complete call."""
    vortex = xspress
    vortex.acquire.sim_put(1)
    status = vortex.complete()
    time.sleep(0.01)
    assert vortex.acquire.get(use_monitor=False) == 0
    assert status.done


def test_collect_xspress(xspress):
    """Check the Xspress3 collects data during fly-scanning."""
    vortex = xspress
    # Kick off the detector
    status = vortex.kickoff()
    vortex.detector_state.sim_put(vortex.detector_states.ACQUIRE)
    status.wait(timeout=3)
    # Set some data so we have something to report
    roi0 = vortex.mcas.mca0.rois.roi0
    roi1 = vortex.mcas.mca1.rois.roi0
    # First data point
    vortex.cam.array_counter.sim_put(1)
    roi0.count.sim_put(280)
    roi1.count.sim_put(216)
    # Second data point
    time.sleep(0.1)  # Simulate the frame being acquired for 0.1 seconds
    vortex.cam.array_counter.sim_put(2)
    roi0.count.sim_put(281)
    roi1.count.sim_put(217)
    assert vortex._fly_data[roi0.count][1][1] == 281
    assert vortex._fly_data[roi1.count][1][1] == 217
    assert vortex._fly_data[vortex.roi_sums.roi0.count][-1][1] == 281 + 217
    # Get data and check its structure
    data = list(vortex.collect())
    datum = data[0]
    assert datum["data"][vortex.mcas.mca0.rois.roi0.count.name] == 281
    assert datum["data"][vortex.mcas.mca1.rois.roi0.count.name] == 217
    assert datum["data"][vortex.roi_sums.roi0.count.name] == 281 + 217
    assert type(datum["time"]) is float


def test_fly_data_xspress(xspress):
    """Check that the Xspress3 processes fly-scanning data."""
    vortex = xspress
    # Set come incomplete fly-scan data
    vortex._fly_data = {
        vortex.cam.array_counter: [
            # From last run, read during ``subscribe(read=True)``
            (51.0, 22),
            # (timestamp, value)
            (100.1, 2),
            (100.2, 3),
            (100.3, 4),
        ],
        vortex.mcas.mca0.rois.roi0.net_count: [
            # From last run, read during ``subscribe(read=True)``
            (51.01, 0),
            # (timestamp, value)
            (100.11, 500),
            (100.21, 498),
            (100.31, 502),
        ],
        vortex.mcas.mca1.rois.roi0.net_count: [
            # From last run, read during ``subscribe(read=True)``
            (51.02, 0),
            # (timestamp, value)
            (100.12, 12.3),
            # This value is omitted to test for filling in missing values
            # (100.22, 14.4),
            (100.32, 9.84),
        ],
        # Make the ROI sums the same as how they get updated from real data
        vortex.roi_sums.roi0: [
            # From last run, read during ``subscribe(read=True)``
            (51.02, 0),
            # (timestamp, value)
            (100.11, 500 + 0),
            (100.12, 500 + 12.3),
            (100.21, 498 + 12.3),
            (100.31, 502 + 12.3),
            (100.32, 502 + 9.84),
        ],
    }
    # Check the process dataframe
    fly_data, fly_ts = vortex.fly_data()
    expected_columns = [
        "timestamps",
        vortex.cam.array_counter,
        vortex.mcas.mca0.rois.roi0.net_count,
        vortex.mcas.mca1.rois.roi0.net_count,
        vortex.roi_sums.roi0,
    ]
    assert list(fly_data.columns) == expected_columns
    assert list(fly_ts.columns) == expected_columns
    # Check that it fills in missing data
    series = fly_data[vortex.mcas.mca1.rois.roi0.net_count]
    np.testing.assert_equal(series.values, [12.3, 12.3, 9.84])
    assert not fly_data.isnull().values.any()
    # Check that it fills in missing timestamps
    series = fly_ts[vortex.mcas.mca1.rois.roi0.net_count]
    np.testing.assert_equal(series.values, [100.12, 100.21, 100.32])
    assert not fly_ts.isnull().values.any()
    # Check that ROI sums are included properly
    mca_sum = fly_data[vortex.roi_sums.roi0]
    np.testing.assert_equal(mca_sum.values, [512.3, 510.3, 511.84])


def test_describe_collect_xspress(xspress):
    vortex = xspress
    # Force all the ROI counts to update
    for mca_num, mca in enumerate(vortex.mca_records()):
        for roi_num in range(vortex.num_rois):
            roi = vortex.get_roi(mca_num, roi_num)
            roi.count.get()
    desc = vortex.describe_collect()
    # Perform some spot-checks for descriptions
    assert vortex.name in desc.keys()
    sub_desc = desc[vortex.name]
    assert vortex.mcas.mca0.total_count.name in sub_desc.keys()
    assert vortex.mcas.mca0.dead_time_percent.name in sub_desc.keys()
    assert vortex.mcas.mca0.spectrum.name in sub_desc.keys()
    assert vortex.mcas.mca0.rois.roi0.net_count.name in sub_desc.keys()
    assert vortex.mcas.mca0.rois.roi0.count.name in sub_desc.keys()


@pytest.mark.skip(reason="DXP fly-scanning not yet implemented")
def test_parse_dxp_buffer(dxp):
    """The output for fly-scanning with the DXP-based readout electronics
    is a raw uint16 buffer that must be parsed by the ophyd device
    according to section 5.3.3 of
    https://cars9.uchicago.edu/software/epics/XMAP_User_Manual.pdf

    """
    vortex = dxp
    fp = Path(__file__)
    buff = np.loadtxt(fp.parent / "dxp_3px_4elem_Fe55.txt")
    data = parse_xmap_buffer(buff)
    assert isinstance(data, dict)
    assert data["num_pixels"] == 3
    assert len(data["pixels"]) == 3


@pytest.mark.parametrize("vortex", DETECTORS, indirect=True)
def test_device_sums(vortex):
    """Does the device correctly calculate the overall counts, etc."""
    assert isinstance(vortex.total_count, Signal)
    spectrum = np.arange(256)
    vortex.mcas.mca0.spectrum.sim_put(spectrum)
    expected = spectrum.sum()
    assert vortex.total_count.get() == expected
    # Add a second spectrum
    spectrum_2 = np.arange(256, 512)
    vortex.mcas.mca1.spectrum.sim_put(spectrum_2)
    expected += spectrum_2.sum()
    assert vortex.total_count.get() == expected


@pytest.mark.parametrize("vortex", DETECTORS, indirect=True)
def test_roi_sums(vortex):
    """Check that we get the sum over all elements for an ROI."""
    # Check that the ROI calc signals exist
    assert isinstance(vortex.roi_sums.roi0, OphydObject)
    # Set some fake ROI values
    vortex.mcas.mca0.rois.roi0.count.sim_put(5)
    assert vortex.roi_sums.roi0.count.get() == 5
    vortex.mcas.mca0.rois.roi0.net_count.sim_put(13)
    assert vortex.roi_sums.roi0.net_count.get() == 13


@pytest.mark.parametrize("vortex", DETECTORS, indirect=True)
def test_mca_calcs(vortex):
    # Check that the ROI calc signals exist
    assert isinstance(vortex.mcas.mca0.total_count, OphydObject)
    # Does it sum together the total counts?
    spectrum = np.random.randint(2**16, size=(vortex.num_rois))
    mca = vortex.mcas.mca0
    mca.spectrum.sim_put(spectrum)
    assert mca.total_count.get(use_monitor=False) == np.sum(spectrum)


@pytest.mark.parametrize("vortex", ["xspress"], indirect=True)
def test_dead_time_calc(vortex):
    assert vortex.dead_time_average.get(use_monitor=False) == 0
    assert vortex.dead_time_max.get(use_monitor=False) == 0
    assert vortex.dead_time_min.get(use_monitor=False) == 0
    # Set the per-element dead-times
    dead_times = [3, 4, 5, 6]
    for mca, dt in zip(vortex.mca_records(), dead_times):
        mca.dead_time_percent.sim_put(dt)
    # Check that the stats get updated
    assert vortex.dead_time_min.get(use_monitor=False) == 3
    assert vortex.dead_time_max.get(use_monitor=False) == 6
    assert vortex.dead_time_average.get(use_monitor=False) == 4.5


def test_default_time_signal_dxp(dxp):
    assert dxp.default_time_signal is dxp.preset_real_time


def test_default_time_signal_xspress(xspress):
    # assert xspress.default_time_signal is xspress.acquire_time
    assert xspress.default_time_signal is xspress.cam.acquire_time


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import numpy as np

__all__ = ["time_converter"]


def time_converter(total_seconds):
    """
    Convert time (in seconds) to a tuple of hours, minutes, seconds

    Parameters
    ==========
    total_seconds (float/int)

    Returns
    ==========
    tuple: hours, minutes, seconds
    """
    if np.isnan(total_seconds):
        return float("nan"), float("nan"), float("nan")
    # Use np.floor to round to the smaller integer
    hours = np.floor(total_seconds / 3600).astype(int)
    minutes = np.floor((total_seconds % 3600) / 60).astype(int)
    seconds = np.round(total_seconds % 60, decimals=1)

    # Convert float values to integers
    return hours, minutes, seconds


def is_valid_value(value):
    """
    Check if the value is considered valid for inclusion in metadata.
    Valid values are non-None, and if they are str/list/tuple/dict, they should have a positive length.

    Parameters
    ==========
        value (any): The value to check.

    Returns
    ==========
        bool: True if the value is valid, False otherwise.
    """
    # Check if the value is None
    if value is None:
        return False
    # Check if the value is a collection with length
    if isinstance(value, (str, list, tuple, dict)):
        return len(value) > 0
    # All other non-None values are considered valid
    return True


# -----------------------------------------------------------------------------
# :author:    Juanjuan Huang
# :email:     juanjuan.huang@anl.gov
# :copyright: Copyright © 2024, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import importlib
import os
from pathlib import Path

from haven import _iconfig
from haven._iconfig import beamline_connected, load_config, print_config_value


def test_default_values():
    config = load_config()
    assert "beamline" in config.keys()


def test_loading_a_file():
    test_file = Path(__file__).resolve().parent / "test_iconfig.toml"
    config = load_config(file_paths=(test_file,))
    assert config["beamline"]["pv_prefix"] == "spam"


def test_config_files_from_env():
    # Set the environmental variable with the path to a test TOML file
    test_file = Path(__file__).resolve().parent / "test_iconfig.toml"
    old_env = os.environ["HAVEN_CONFIG_FILES"]
    os.environ["HAVEN_CONFIG_FILES"] = str(test_file)
    importlib.reload(_iconfig)
    try:
        # Load the configuration
        importlib.reload(_iconfig)
        config = _iconfig.load_config()
        # Check that the test file was loaded
        assert config["beamline"]["pv_prefix"] == "spam"
    finally:
        # Reset the old configuration to avoid breaking future tests
        os.environ["HAVEN_CONFIG_FILES"] = old_env
        importlib.reload(_iconfig)


def test_merging_dicts():
    """Do the entries from multiple dictioneries merge properly?"""
    this_dir = Path(__file__).resolve().parent
    default_files = [
        this_dir.parent / "iconfig_testing.toml",
    ]
    test_file = this_dir / "test_iconfig.toml"
    config = load_config(file_paths=(*default_files, test_file))
    print(config)
    assert "prefix" in config["area_detector"][0].keys()


def test_haven_config_cli(capsys):
    """Test the function used as a CLI way to get config values."""
    print_config_value(["xray_source.prefix"])
    # Check stdout for config value
    captured = capsys.readouterr()
    assert captured.out == "ID255ds:\n"


def test_beamline_connected():
    """Check the context manager to temporarily connect the beamline."""
    config = load_config()
    assert not config["beamline"]["hardware_is_present"]
    with beamline_connected():
        config = load_config()
        assert config["beamline"]["hardware_is_present"]
    config = load_config()
    assert not config["beamline"]["hardware_is_present"]


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging

import numpy as np
from bluesky_queueserver_api import BPlan
from qasync import asyncSlot
from qtpy import QtWidgets
from qtpy.QtGui import QDoubleValidator

from firefly.component_selector import ComponentSelector
from firefly.plans import regions_display

log = logging.getLogger()


class TitleRegion:
    def __init__(self):
        self.setup_ui()

    def setup_ui(self):
        self.layout = QtWidgets.QGridLayout()
        labels = [
            "Priority axis",
            "Motor",
            "Start",
            "Stop",
            "Scan points",
            "Snake",
            "Fly",
        ]
        Qlabels_all = {}

        # add labels in the first row
        for i, label_i in enumerate(labels):
            Qlabel_i = QtWidgets.QLabel(label_i)
            self.layout.addWidget(Qlabel_i, 0, i)
            Qlabels_all[label_i] = Qlabel_i

        # fix widths so the labels are aligned with GridScanRegions
        Qlabels_all["Priority axis"].setFixedWidth(70)
        Qlabels_all["Motor"].setFixedWidth(100)
        Qlabels_all["Scan points"].setFixedWidth(68)
        Qlabels_all["Snake"].setFixedWidth(53)
        Qlabels_all["Fly"].setFixedWidth(43)

        # add labels in the second row
        label = QtWidgets.QLabel("fast -> slow")
        self.layout.addWidget(label, 1, 0)


class GridScanRegion(regions_display.RegionBase):

    def setup_ui(self):
        self.layout = QtWidgets.QHBoxLayout()

        # First item, motor No.
        self.motor_label = QtWidgets.QLCDNumber()
        self.motor_label.setStyleSheet(
            "QLCDNumber { background-color: white; color: red; }"
        )
        self.motor_label.display(self.line_label)
        self.layout.addWidget(self.motor_label)

        # Second item, ComponentSelector
        self.motor_box = ComponentSelector()
        self.layout.addWidget(self.motor_box)

        # Third item, start point
        self.start_line_edit = QtWidgets.QLineEdit()
        self.start_line_edit.setValidator(QDoubleValidator())  # only takes floats
        self.start_line_edit.setPlaceholderText("Start…")
        self.layout.addWidget(self.start_line_edit)

        # Forth item, stop point
        self.stop_line_edit = QtWidgets.QLineEdit()
        self.stop_line_edit.setValidator(QDoubleValidator())  # only takes floats
        self.stop_line_edit.setPlaceholderText("Stop…")
        self.layout.addWidget(self.stop_line_edit)

        # Fifth item, number of scan point
        self.scan_pts_spin_box = QtWidgets.QSpinBox()
        self.scan_pts_spin_box.setMinimum(1)
        self.scan_pts_spin_box.setMaximum(99999)
        self.layout.addWidget(self.scan_pts_spin_box)

        # Sixth item, snake checkbox
        self.snake_checkbox = QtWidgets.QCheckBox()
        self.snake_checkbox.setText("Snake")
        self.snake_checkbox.setEnabled(True)
        self.layout.addWidget(self.snake_checkbox)

        # Seventh item, fly checkbox # not available right now
        self.fly_checkbox = QtWidgets.QCheckBox()
        self.fly_checkbox.setText("Fly")
        self.fly_checkbox.setEnabled(False)
        self.layout.addWidget(self.fly_checkbox)


class GridScanDisplay(regions_display.RegionsDisplay):
    Region = GridScanRegion
    default_num_regions = 2

    def __init__(self, parent=None, args=None, macros=None, ui_filename=None, **kwargs):
        super().__init__(parent, args, macros, ui_filename, **kwargs)

    def customize_ui(self):
        super().customize_ui()
        self.update_snakes()
        # add title layout
        self.title_region = TitleRegion()
        self.ui.title_layout.addLayout(self.title_region.layout)
        # When selections of detectors changed update_total_time
        # self.ui.detectors_list.selectionModel().selectionChanged.connect(
        #     self.update_total_time
        # )
        self.ui.spinBox_repeat_scan_num.valueChanged.connect(self.update_total_time)
        # Connect scan points change to update total time
        for region in self.regions:
            region.scan_pts_spin_box.valueChanged.connect(self.update_total_time)

    def time_per_scan(self, detector_time):
        total_num_pnts = np.prod(
            [region_i.scan_pts_spin_box.value() for region_i in self.regions]
        )
        total_time_per_scan = total_num_pnts * detector_time
        return total_time_per_scan

    @asyncSlot(object)
    async def update_devices_slot(self, registry):
        await super().update_devices(registry)
        await self.detectors_list.update_devices(registry)

    @asyncSlot(int)
    async def update_regions_slot(self, new_region_num: int):
        await super().update_regions(new_region_num)
        self.update_snakes()

    def update_snakes(self):
        """Update the snake checkboxes.

        The last region is not snakable, so that checkbox gets
        disabled and unchecked. The rest get enabled.

        """
        if len(self.regions) > 0:
            self.regions[-1].snake_checkbox.setEnabled(False)
            self.regions[-1].snake_checkbox.setChecked(False)
            for region_i in self.regions[:-1]:
                region_i.snake_checkbox.setEnabled(True)

    def get_scan_parameters(self):
        # Get scan parameters from widgets
        detectors = self.ui.detectors_list.selected_detectors()
        repeat_scan_num = int(self.ui.spinBox_repeat_scan_num.value())

        # Get paramters from each rows of line regions:
        motor_lst, start_lst, stop_lst, num_points_lst = [], [], [], []
        for region_i in reversed(self.regions):
            motor_lst.append(region_i.motor_box.current_component().name)
            start_lst.append(float(region_i.start_line_edit.text()))
            stop_lst.append(float(region_i.stop_line_edit.text()))
            num_points_lst.append(int(region_i.scan_pts_spin_box.value()))

        motor_args = [
            values
            for motor_i in zip(motor_lst, start_lst, stop_lst, num_points_lst)
            for values in motor_i
        ]

        return detectors, motor_args, repeat_scan_num

    def queue_plan(self, *args, **kwargs):
        """Execute this plan on the queueserver."""
        detectors, motor_args, repeat_scan_num = self.get_scan_parameters()
        md = self.get_meta_data()

        # get snake axes, if all unchecked, set it None
        snake_axes = [
            region_i.motor_box.current_component().name
            for i, region_i in enumerate(self.regions)
            if region_i.snake_checkbox.isChecked()
        ]

        if snake_axes == []:
            snake_axes = False

        if self.ui.relative_scan_checkbox.isChecked():
            scan_type = "rel_grid_scan"
        else:
            scan_type = "grid_scan"

        # Build the queue item
        item = BPlan(
            scan_type,
            detectors,
            *motor_args,
            snake_axes=snake_axes,
            md=md,
        )

        # Submit the item to the queueserver
        log.info(f"Added grid_scan() plan to queue ({repeat_scan_num} scans).")
        # repeat scans
        for i in range(repeat_scan_num):
            self.queue_item_submitted.emit(item)

    def ui_filename(self):
        return "plans/grid_scan.ui"


# -----------------------------------------------------------------------------
# :author:    Juanjuan Huang
# :email:     juanjuan.huang@anl.gov
# :copyright: Copyright © 2024, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging

import numpy as np
from ophyd import Component as Cpt
from ophyd import Device, EpicsMotor
from ophyd import FormattedComponent as FCpt
from ophyd import PseudoPositioner, PseudoSingle
from ophyd.pseudopos import pseudo_position_argument, real_position_argument

log = logging.getLogger(__name__)

um_per_mm = 1000


class RowlandPositioner(PseudoPositioner):
    """A pseudo positioner describing a rowland circle.

    Real Axes
    =========
    x
    y
    z
    z1

    Pseudo Axes
    ===========
    D
      In mm
    theta
      In degrees
    alpha
      In degrees
    """

    def __init__(
        self,
        x_motor_pv: str,
        y_motor_pv: str,
        z_motor_pv: str,
        z1_motor_pv: str,
        *args,
        **kwargs,
    ):
        self.x_motor_pv = x_motor_pv
        self.y_motor_pv = y_motor_pv
        self.z_motor_pv = z_motor_pv
        self.z1_motor_pv = z1_motor_pv
        super().__init__(*args, **kwargs)

    # Pseudo axes
    D: PseudoSingle = Cpt(PseudoSingle, name="D", limits=(0, 1000))
    theta: PseudoSingle = Cpt(PseudoSingle, name="theta", limits=(0, 180))
    alpha: PseudoSingle = Cpt(PseudoSingle, name="alpha", limits=(0, 180))

    # Real axes
    x: EpicsMotor = FCpt(EpicsMotor, "{x_motor_pv}", name="x")
    y: EpicsMotor = FCpt(EpicsMotor, "{y_motor_pv}", name="y")
    z: EpicsMotor = FCpt(EpicsMotor, "{z_motor_pv}", name="z")
    z1: EpicsMotor = FCpt(EpicsMotor, "{z1_motor_pv}", name="z1")

    @pseudo_position_argument
    def forward(self, pseudo_pos):
        """Run a forward (pseudo -> real) calculation"""
        # Convert distance to microns and degrees to radians
        D = pseudo_pos.D * um_per_mm
        theta = pseudo_pos.theta / 180.0 * np.pi
        alpha = pseudo_pos.alpha / 180.0 * np.pi
        # Convert virtual positions to real positions
        x = D * (np.sin(theta + alpha)) ** 2
        y = D * ((np.sin(theta + alpha)) ** 2 - (np.sin(theta - alpha)) ** 2)
        z1 = D * np.sin(theta - alpha) * np.cos(theta + alpha)
        z2 = D * np.sin(theta - alpha) * np.cos(theta - alpha)
        z = z1 + z2
        print(x, y, z1, z)
        return self.RealPosition(
            x=x,
            y=y,
            z=z,
            z1=z1,
        )

    @real_position_argument
    def inverse(self, real_pos):
        """Run an inverse (real -> pseudo) calculation"""
        return self.PseudoPosition(D=0, theta=0, alpha=0)
        # Expand the variables
        x = real_pos.x
        y = real_pos.y
        z = real_pos.z
        z1 = real_pos.z1
        # Invert the calculation, first for 'd'
        a = y
        b = -x
        c = z1**2
        d = -(z1**2 * y)  # Maybe needs parens?
        p = (3 * a * c - b**2) / (3 * a**2)
        q = (2 * b**3 - 9 * a * b * c + 27 * a**2 * d) / (27 * a**3)
        D = (-(q / 2) + (q**2 / 4 + p**3 / 27) ** 0.5) ** (1 / 3) + (
            -(q / 2) - (q**2 / 4 + p**3 / 27) ** 0.5
        ) ** (1 / 3)
        # D = x / ((1 - (z1 ** 2 / (D ** 2 * y + z1 ** 2))) ** 2)
        # cos(theta + alpha) = (z1 ** 2 / (D ** 2 * y + z1 ** 2)) ** 0.5
        print(a, b, c, d)
        print(p, q, D)
        return self.PseudoPosition(D=D, theta=0, alpha=0)


# Rewrite the following four equations so that they calculate D, theta and alpha based on inputting x, y, z1 and z.

#         x = D * (sin(theta + alpha)) ** 2
#         y = D * ((sin(theta + alpha)) ** 2 - (sin(theta - alpha)) ** 2)
#         z1 = D * sin(theta - alpha) * cos(theta + alpha)
#         z = z1 + D * sin(theta - alpha) * cos(theta - alpha)

# To calculate D, theta, and alpha based on x, y, z1, and z, we can rearrange the given equations as follows:

# D = x / (sin(theta + alpha)) ** 2

# sin(theta - alpha) = z1 / (D * cos(theta + alpha))

# cos(theta - alpha) = (z - z1) / (D * sin(theta - alpha))

# Substitute the above two equations into the expression for y:

# y = D * ((sin(theta + alpha)) ** 2 - (sin(theta - alpha)) ** 2)

# y = D * ((sin(theta + alpha)) ** 2 - (z1 / (D * cos(theta + alpha))) ** 2)

# y = D * ((sin(theta + alpha)) ** 2 - (z1 ** 2 / D ** 2) * (cos(theta + alpha)) ** (-2))

# Solve for cos(theta + alpha) in the above equation:

# cos(theta + alpha) = (z1 ** 2 / (D ** 2 * y + z1 ** 2)) ** 0.5

# Substitute the value of cos(theta + alpha) into the equation for sin(theta - alpha):

# sin(theta - alpha) = z1 / (D * (z1 ** 2 / (D ** 2 * y + z1 ** 2)) ** 0.5)

# Simplify the above equation:

# sin(theta - alpha) = (z1 ** 2 / (D ** 2 * y + z1 ** 2)) ** 0.5

# Finally, substitute the values of cos(theta + alpha) and sin(theta - alpha) into the equation for D:

# D = x / ((sin(theta + alpha)) ** 2)

# D = x / ((1 - (z1 ** 2 / (D ** 2 * y + z1 ** 2))) ** 2)

# This equation can be solved numerically to obtain the value of D. Once D is known, we can use the equations for cos(theta + alpha) and sin(theta - alpha) to calculate theta and alpha.


class LERIXSpectrometer(Device):
    rowland = Cpt(
        RowlandPositioner,
        x_motor_pv="vme_crate_ioc:m1",
        y_motor_pv="vme_crate_ioc:m2",
        z_motor_pv="vme_crate_ioc:m3",
        z1_motor_pv="vme_crate_ioc:m4",
        name="rowland",
    )


# async def make_lerix_device(name: str, x_pv: str, y_pv: str, z_pv: str, z1_pv: str):
#     dev = RowlandPositioner(
#         name=name,
#         x_motor_pv=x_pv,
#         y_motor_pv=y_pv,
#         z_motor_pv=z_pv,
#         z1_motor_pv=z1_pv,
#         labels={"lerix_spectrometers"},
#     )
#     pvs = ", ".join((x_pv, y_pv, z_pv, z1_pv))
#     try:
#         await await_for_connection(dev)
#     except TimeoutError as exc:
#         log.warning(f"Could not connect to LERIX spectrometer: {name} ({pvs})")
#     else:
#         log.info(f"Created area detector: {name} ({pvs})")
#         return dev


# def load_lerix_spectrometers(config=None):
#     """Create devices for the LERIX spectrometer."""
#     if config is None:
#         config = load_config()
#     # Create spectrometers
#     devices = []
#     for name, cfg in config.get("lerix", {}).items():
#         rowland = cfg["rowland"]
#         devices.append(
#             make_device(
#                 RowlandPositioner,
#                 name=name,
#                 x_motor_pv=rowland["x_motor_pv"],
#                 y_motor_pv=rowland["y_motor_pv"],
#                 z_motor_pv=rowland["z_motor_pv"],
#                 z1_motor_pv=rowland["z1_motor_pv"],
#                 labels={"lerix_spectromoters"},
#             )
#         )
#     return devices


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import gc

import databroker
import pytest
from bluesky import RunEngine
from ophyd.sim import instantiate_fake_device

from haven import run_engine
from haven.devices.aps import ApsMachine


@pytest.fixture()
def aps(sim_registry):
    aps = instantiate_fake_device(ApsMachine, name="advanced_photon_source")


def test_subscribers_garbage_collection(monkeypatch, aps):
    """Tests for regression of a bug in databroker.

    Since databroker uses a weak reference to the insert function, it
    can be subject to garbage collection and no longer able to save
    data.

    """
    monkeypatch.setattr(databroker, "catalog", {"bluesky": databroker.temp()})
    RE = run_engine(use_bec=False)
    assert len(RE.dispatcher.cb_registry.callbacks) == 12
    gc.collect()
    assert len(RE.dispatcher.cb_registry.callbacks) == 12


def test_run_engine_preprocessors(aps):
    RE = run_engine(use_bec=False)
    assert len(RE.preprocessors) > 0


def test_run_engine_created(aps):
    RE = run_engine(use_bec=False)
    assert isinstance(RE, RunEngine)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import pytest
from ophyd.sim import make_fake_device

from firefly.kb_mirrors import KBMirrorsDisplay
from haven.devices.mirrors import KBMirrors


@pytest.fixture()
async def kb_mirrors(sim_registry):
    """A fake set of slits using the 4-blade setup."""
    kb = KBMirrors(
        prefix="255idc:KB:",
        name="kb_mirrors",
        horiz_upstream_motor="255idc:m1",
        vert_upstream_motor="255idc:m2",
        horiz_downstream_motor="255idc:m3",
        vert_downstream_motor="255idc:m4",
    )
    return kb


@pytest.fixture()
def kb_bendable_mirrors(sim_registry):
    """A fake set of slits using the 4-blade setup."""
    FakeMirrors = make_fake_device(KBMirrors)
    kb = FakeMirrors(
        prefix="255idc:Long_KB:",
        name="kb_bendable_mirrors",
        horiz_upstream_motor="255idc:m5",
        vert_upstream_motor="255idc:m6",
        horiz_downstream_motor="255idc:m7",
        vert_downstream_motor="255idc:m8",
        horiz_upstream_bender="255idc:m21",
        vert_upstream_bender="255idc:m22",
        horiz_downstream_bender="255idc:m23",
        vert_downstream_bender="255idc:m24",
    )
    sim_registry.register(kb)
    return kb


@pytest.fixture()
def display(kb_mirrors, qtbot):
    disp = KBMirrorsDisplay(macros={"DEVICE": kb_mirrors.name})
    qtbot.addWidget(disp)
    return disp


def test_bender_widgets(qtbot, kb_bendable_mirrors):
    display = KBMirrorsDisplay(macros={"DEVICE": kb_bendable_mirrors.name})
    qtbot.addWidget(display)
    # Check that the bender control widgets were enabled
    assert display.ui.horizontal_upstream_display.isEnabled()
    assert display.ui.horizontal_downstream_display.isEnabled()
    assert display.ui.vertical_upstream_display.isEnabled()
    assert display.ui.vertical_downstream_display.isEnabled()


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import asyncio

import pytest
from ophyd_async.core import get_mock_put, set_mock_value

from haven.devices.xray_source import BusyStatus, PlanarUndulator


@pytest.fixture()
async def undulator():
    undulator = PlanarUndulator(prefix="PSS:255ID:", name="undulator")
    await undulator.connect(mock=True)
    return undulator


async def test_set_energy(undulator):
    # Set the energy
    status = undulator.energy.set(5)
    # Fake the done PV getting updated
    set_mock_value(undulator.energy.done, BusyStatus.BUSY)
    await asyncio.sleep(0.01)  # Let the event loop run
    set_mock_value(undulator.energy.done, BusyStatus.DONE)
    # Check that the signals got set properly
    await status
    assert await undulator.energy.setpoint.get_value() == 5


async def test_stop_energy(undulator):
    stop_mock = get_mock_put(undulator.stop_button)
    assert not stop_mock.called
    await undulator.energy.stop()
    assert stop_mock.called


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging
import time
import warnings
from typing import Mapping, Optional

from bluesky_queueserver_api import comm_base
from bluesky_queueserver_api.zmq.aio import REManagerAPI
from qasync import asyncSlot
from qtpy.QtCore import QObject, QTimer, Signal

from haven import load_config
from haven.exceptions import InvalidConfiguration

log = logging.getLogger()


def queueserver_api():
    try:
        config = load_config()["queueserver"]
        ctrl_addr = f"tcp://{config['control_host']}:{config['control_port']}"
        info_addr = f"tcp://{config['info_host']}:{config['info_port']}"
    except KeyError as e:
        raise InvalidConfiguration(str(e))
    api = REManagerAPI(zmq_control_addr=ctrl_addr, zmq_info_addr=info_addr)
    return api


class QueueClient(QObject):
    api: REManagerAPI
    _last_queue_status: Optional[dict] = None
    last_update: float = -1
    timeout: float = 0.2
    min_timeout: float = 0.2
    timer: QTimer

    # Signals responding to queue changes
    status_changed = Signal(dict)
    length_changed = Signal(int)
    in_use_changed = Signal(bool)  # If length > 0, or queue is running
    autostart_changed = Signal(bool)
    queue_stop_changed = Signal(bool)  # If a queue stop has been requested
    environment_opened = Signal(bool)  # Opened (True) or closed (False)
    environment_state_changed = Signal(str)  # New state
    manager_state_changed = Signal(str)  # New state
    re_state_changed = Signal(str)  # New state
    devices_changed = Signal(dict)

    def start(self):
        self.timer.start(int(self.timeout * 1000))

    def __init__(self, *args, api, **kwargs):
        self.api = api
        super().__init__(*args, **kwargs)
        self._last_queue_status = {}
        # Setup timer for updating the queue
        self.timer = QTimer()
        self.timer.timeout.connect(self.update)

    @asyncSlot(bool)
    async def open_environment(self, to_open):
        if to_open:
            result = await self.api.environment_open()
        else:
            result = await self.api.environment_close()
        if result["success"]:
            self.environment_opened.emit(to_open)
        else:
            log.error(f"Failed to open/close environment: {result['msg']}")

    @asyncSlot()
    async def update(self):
        now = time.monotonic()
        if now >= self.last_update + self.timeout:
            if not load_config()["beamline"]["hardware_is_present"]:
                log.warning("Beamline not connected, skipping queue client update.")
                self.timeout = 60  # Just update every 1 minute
                self.last_update = now
                return
            log.debug("Updating queue client.")
            try:
                await self._check_queue_status()
            except comm_base.RequestTimeoutError as e:
                # If we can't reach the server, wait for a minute and retry
                self.timeout = min(60, self.timeout * 2)
                log.warn(str(e))
                warnings.warn(str(e))
                log.info(f"Retrying in {self.timeout} seconds.")
            else:
                # Update succeeded, so wait for a bit
                self.timeout = self.min_timeout
            finally:
                self.last_update = now

    @asyncSlot(bool)
    async def request_pause(self, *args, defer: bool = True, **kwargs):
        """Ask the queueserver run engine to pause.

        Parameters
        ==========
        defer
          If true, the run engine will be paused at the next
          checkpoint. Otherwise, it will pause now.

        """
        option = "deferred" if defer else "immediate"
        await self.api.re_pause(option=option)

    @asyncSlot(object)
    async def add_queue_item(self, item):
        log.info(f"Client adding item to queue: {item}")
        try:
            result = await self.api.item_add(item=item)
            self.check_result(result)
        except (RuntimeError, comm_base.RequestFailedError) as ex:
            # Request failed, so force a UI update
            await self.check_queue_status(force=True)
            raise
        else:
            await self.check_queue_status(force=False)

    @asyncSlot(bool)
    async def toggle_autostart(self, enable: bool):
        log.debug(f"Toggling auto-start: {enable}")
        try:
            result = await self.api.queue_autostart(enable)
            self.check_result(result, task="toggle auto-start")
        except (RuntimeError, comm_base.RequestFailedError):
            # Request failed, so force a UI update
            await self.check_queue_status(force=True)
        else:
            await self.check_queue_status(force=False)

    @asyncSlot(bool)
    async def stop_queue(self, stop: bool):
        """Turn on/off whether the queue will stop after the current plan."""
        # Determine which call to usee
        if stop:
            api_call = self.api.queue_stop()
        else:
            api_call = self.api.queue_stop_cancel()
        # Execute the call
        try:
            result = await api_call
            self.check_result(result, task="toggle stop queue")
        except (RuntimeError, comm_base.RequestFailedError):
            # Request failed, so force a UI update
            await self.check_queue_status(force=True)
        else:
            await self.check_queue_status(force=False)

    @asyncSlot()
    async def start_queue(self):
        result = await self.api.queue_start()
        self.check_result(result, task="start queue")

    @asyncSlot()
    async def resume_runengine(self):
        result = await self.api.re_resume()
        self.check_result(result, task="resume run engine")

    @asyncSlot()
    async def stop_runengine(self):
        result = await self.api.re_stop()
        self.check_result(result, task="stop run engine")

    @asyncSlot()
    async def abort_runengine(self):
        result = await self.api.re_abort()
        self.check_result(result, task="abort run engine")

    @asyncSlot()
    async def halt_runengine(self):
        result = await self.api.re_halt()
        self.check_result(result, task="halt run engine")

    def check_result(self, result: Mapping, task: str = "control queue server"):
        """Send the result of an API call to the correct logger.

        Expects *result* to have at least the "success" key.

        """
        # Report results
        if result["success"] is True:
            log.debug(f"{task}: {result}")
        else:
            msg = f"Failed to {task}: {result}"
            log.error(msg)
            raise RuntimeError(msg)

    @asyncSlot()
    async def check_queue_status(self, force=False, *args, **kwargs):
        """Get an update queue status from queue server and notify slots.

        Parameters
        ==========
        force
          If false (default), the ``queue_status_changed`` signal will
          be emitted only if the queue status has changed since last
          check.
        *args, *kwargs
          Unused arguments from qt widgets

        Emits
        =====
        self.status_changed
          With the updated queue status.

        """
        try:
            await self._check_queue_status(force=force)
        except comm_base.RequestTimeoutError as e:
            log.warning(str(e))
            warnings.warn(str(e))

    async def _check_queue_status(self, force: bool = False):
        """Get an update queue status from queue server and notify slots.

        Similar to ``check_queue_status`` but without the exception
        handling.

        Parameters
        ==========
        force
          If false (default), the ``queue_status_changed`` signal will
          be emitted only if the queue status has changed since last
          check.

        Emits
        =====
        self.status_changed
          With the updated queue status.

        """
        new_status = await self.api.status()
        # Add a new key for whether the queue is busy (length > 0 or running)
        has_queue = new_status["items_in_queue"] > 0
        is_running = new_status["manager_state"] in [
            "paused",
            "starting_queue",
            "executing_queue",
            "executing_task",
        ]
        new_status.setdefault("in_use", has_queue or is_running)
        # Check individual components of the status if they've changed
        signals_to_check = [
            # (status key, signal to emit)
            ("worker_environment_exists", self.environment_opened),
            ("worker_environment_state", self.environment_state_changed),
            ("manager_state", self.manager_state_changed),
            ("re_state", self.re_state_changed),
            ("items_in_queue", self.length_changed),
            ("in_use", self.in_use_changed),
            ("queue_stop_pending", self.queue_stop_changed),
            ("queue_autostart_enabled", self.autostart_changed),
        ]
        if force:
            log.debug(f"Forcing queue server status update: {new_status}")
        for key, signal in signals_to_check:
            is_new = key not in self._last_queue_status
            has_changed = new_status[key] != self._last_queue_status.get(key)
            if is_new or has_changed or force:
                signal.emit(new_status[key])
        # Check for new available devices
        if new_status["devices_allowed_uid"] != self._last_queue_status.get(
            "devices_allowed_uid"
        ):
            await self.update_devices()
        # check the whole status to see if it's changed
        has_changed = new_status != self._last_queue_status
        if has_changed or force:
            self.status_changed.emit(new_status)
            self._last_queue_status = new_status

    async def update_devices(self):
        "Emit the latest dict of available devices."
        response = await self.api.devices_allowed()
        if response["success"]:
            devices = response["devices_allowed"]
            self.devices_changed.emit(devices)
        else:
            log.warning(
                "Could not poll devices_allowed:"
                f" {response.get('msg', 'reason unknown.')}"
            )


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import numpy as np
import pytest
from ophyd import sim

from haven import KRange, energy_scan, xafs_scan


@pytest.fixture()
def mono_motor():
    return sim.SynAxis(name="mono_energy", labels={"motors", "energies"})


@pytest.fixture()
def id_gap_motor():
    return sim.SynAxis(name="id_gap_energy", labels={"motors", "energies"})


@pytest.fixture()
def exposure_motor():
    return sim.Signal(name="exposure")


@pytest.fixture()
def energies():
    return np.linspace(8300, 8500, 100)


@pytest.fixture()
def I0(sim_registry):
    # Register an ion chamber
    I0 = sim.SynGauss(
        "I0",
        sim.motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={"ion_chambers"},
    )
    sim_registry.register(I0)
    return I0


def test_energy_scan_basics(
    beamline_manager, mono_motor, id_gap_motor, energies, RE, tmp_path
):
    beamline_manager.local_storage.full_path._readback = str(tmp_path)
    exposure_time = 1e-3
    # Set up fake detectors and motors
    I0_exposure = sim.SynAxis(
        name="I0_exposure",
        labels={
            "exposures",
        },
    )
    It_exposure = sim.SynAxis(
        name="It_exposure",
        labels={
            "exposures",
        },
    )
    I0 = sim.SynGauss(
        name="I0",
        motor=mono_motor,
        motor_field="mono_energy",
        center=np.median(energies),
        Imax=1,
        sigma=1,
        labels={"detectors"},
    )
    It = sim.SynSignal(
        func=lambda: 1.0,
        name="It",
        exposure_time=exposure_time,
    )
    # Execute the plan
    scan = energy_scan(
        energies,
        detectors=[I0, It],
        exposure=exposure_time,
        energy_signals=[mono_motor, id_gap_motor],
        time_signals=[I0_exposure, It_exposure],
        md={"edge": "Ni_K"},
    )
    result = RE(scan, sample_name="xafs_sample")
    # Check that the mono and ID gap ended up in the right position
    # time.sleep(1.0)
    assert mono_motor.readback.get() == np.max(energies)
    # assert mono_motor.get().readback == np.max(energies)
    assert id_gap_motor.get().readback == np.max(energies)
    assert I0_exposure.get().readback == exposure_time
    assert It_exposure.get().readback == exposure_time


def test_raises_on_empty_positioners(RE, energies):
    with pytest.raises(ValueError):
        RE(energy_scan(energies, energy_signals=[]))


def test_single_range(mono_motor, exposure_motor, I0):
    E0 = 10000
    expected_energies = np.arange(9990, 10001, step=1)
    expected_exposures = np.asarray([1.0])
    scan = xafs_scan(
        -10,
        1,
        1,
        0,
        E0=E0,
        energy_signals=[mono_motor],
        time_signals=[exposure_motor],
    )
    # Check that the mono motor is moved to the correct positions
    scan_list = list(scan)
    real_energies = [
        i.args[0] for i in scan_list if i[0] == "set" and i.obj.name == "mono_energy"
    ]
    np.testing.assert_equal(real_energies, expected_energies)
    # Check that the exposure is set correctly
    real_exposures = [
        i.args[0] for i in scan_list if i[0] == "set" and i.obj.name == "exposure"
    ]
    np.testing.assert_equal(real_exposures, expected_exposures)


def test_multi_range(mono_motor, exposure_motor, I0):
    E0 = 10000
    expected_energies = np.concatenate(
        [
            np.arange(9990, 10001, step=2),
            np.arange(10001, 10011, step=1),
        ]
    )
    expected_exposures = np.asarray([0.5, 1.0])
    scan = xafs_scan(
        -10,
        2,
        0.5,
        0,
        1,
        1.0,
        10,
        E0=E0,
        energy_signals=[mono_motor],
        time_signals=[exposure_motor],
    )
    # Check that the mono motor is moved to the correct positions
    scan_list = list(scan)
    real_energies = [
        i.args[0] for i in scan_list if i[0] == "set" and i.obj.name == "mono_energy"
    ]
    np.testing.assert_equal(real_energies, expected_energies)
    # Check that the exposure is set correctly
    real_exposures = [
        i.args[0] for i in scan_list if i[0] == "set" and i.obj.name == "exposure"
    ]
    np.testing.assert_equal(real_exposures, expected_exposures)


def test_exafs_k_range(mono_motor, exposure_motor, I0):
    """Ensure that passing in k_min, etc. produces an energy range
    in K-space.

    """
    E0 = 10000
    E_min = 10
    k_min = 1.6200877248145786
    krange = KRange(k_min=k_min, k_max=14, k_step=0.5, k_weight=0.5, exposure=0.75)
    expected_energies = krange.energies() + E0
    expected_exposures = krange.exposures()
    scan = xafs_scan(
        E_min=10,
        k_step=0.5,
        k_max=14,
        k_exposure=0.75,
        k_weight=0.5,
        E0=E0,
        energy_signals=[mono_motor],
        time_signals=[exposure_motor],
    )
    # Check that the mono motor is moved to the correct positions
    scan_list = list(scan)
    real_energies = [
        i.args[0] for i in scan_list if i[0] == "set" and i.obj.name == "mono_energy"
    ]
    np.testing.assert_equal(real_energies, expected_energies)
    # Check that the exposure is set correctly
    real_exposures = [
        i.args[0] for i in scan_list if i[0] == "set" and i.obj.name == "exposure"
    ]
    np.testing.assert_equal(real_exposures, expected_exposures)


def test_named_E0(mono_motor, exposure_motor, I0):
    expected_energies = np.concatenate(
        [
            np.arange(8323, 8334, step=2),
            np.arange(8334, 8344, step=1),
        ]
    )
    expected_exposures = np.asarray([0.5, 1.0])
    scan = xafs_scan(
        -10,
        2,
        0.5,
        0,
        1,
        1.0,
        10,
        E0="Ni_K",
        energy_signals=[mono_motor],
        time_signals=[exposure_motor],
    )
    # Check that the mono motor is moved to the correct positions
    scan_list = list(scan)
    real_energies = [
        i.args[0] for i in scan_list if i[0] == "set" and i.obj.name == "mono_energy"
    ]
    np.testing.assert_equal(real_energies, expected_energies)
    # Check that the exposure is set correctly
    real_exposures = [
        i.args[0] for i in scan_list if i[0] == "set" and i.obj.name == "exposure"
    ]
    np.testing.assert_equal(real_exposures, expected_exposures)


def test_uses_default_time_signals(dxp, mono_motor):
    """Test that the default time positioners are used if no specific ones are given."""
    scan = xafs_scan(
        -10, 2, 0.5, 0, detectors=[dxp], time_signals=None, energy_signals=[mono_motor]
    )
    msgs = list(scan)
    set_msgs = [m for m in msgs if m.command == "set" and dxp.name in m.obj.name]
    assert len(set_msgs) == 1
    time_msg = set_msgs[0]
    assert time_msg.obj is dxp.preset_real_time
    assert time_msg.args[0] == 0.5


def test_remove_duplicate_energies(mono_motor, exposure_motor, I0):
    plan = xafs_scan(
        -4,
        2,
        1.0,
        6,
        34,
        1.0,
        40,
        E0=8333,
        energy_signals=[mono_motor],
        time_signals=[exposure_motor],
    )
    msgs = list(plan)
    set_msgs = [m for m in msgs if m.command == "set" and m.obj.name == "mono_energy"]
    read_msgs = [m for m in msgs if m.command == "read" and m.obj.name == "I0"]
    energies = [m.args[0] for m in set_msgs]
    # Make sure we only read each point once
    assert len(read_msgs) == len(energies)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from unittest import mock

import pytest

from firefly.ion_chamber import IonChamberDisplay


@pytest.fixture()
async def display(ion_chamber, qtbot):
    display = IonChamberDisplay(macros={"IC": ion_chamber.name})
    qtbot.addWidget(display)
    display.launch_caqtdm = mock.MagicMock()
    return display


async def test_display(display):
    pass



"""A bluesky plan to scan the X-ray energy and capture detector signals.

"""

import logging
from collections import ChainMap
from typing import Mapping, Optional, Sequence, Union

import numpy as np
from bluesky import plans as bp

from .._iconfig import load_config
from ..constants import edge_energy
from ..instrument import beamline
from ..preprocessors import baseline_decorator
from ..typing import DetectorList

__all__ = ["energy_scan"]


log = logging.getLogger(__name__)


# @shutter_suspend_decorator()
@baseline_decorator()
def energy_scan(
    energies: Sequence[float],
    exposure: Union[float, Sequence[float]] = 0.1,
    E0: Union[float, str] = 0,
    detectors: DetectorList = "ion_chambers",
    energy_signals: Sequence = ["energy"],
    time_signals: Optional[Sequence] = None,
    md: Mapping = {},
):
    """Collect a spectrum by scanning X-ray energy.

    For scanning over a pre-defined X-ray absorption edge, try
    :py:func:`~haven.plans.xafs_scan.xafs_scan` instead.

    *exposure* can be either a float, or a sequence of floats. If a
    single value is provided, it will be used for all energies. If a
    sequence is provided, it should be the same length as *energies*
    and the each entry will be used for the corresponding entry in
    *energies*.

    The calculated exposure times will be set for every signal in
    *time_signals*. If *time_signals* is ``None``, then
    *time_signals* will be determined automatically from
    *detectors*: for each detector, if it has an attribute/property
    *default_time_signal*, then this signal will be included in
    *time_signals*.

    **Usage:**

    The following code will run a scan with 1 eV steps from 13000 to
    13099 eV.

        energies = range(13000, 13100)
        RE(energy_scan(energies, exposure=0.5))

    The preparation of energies is up to the calling function. For
    more sophisticated energy lists, consider using the utility
    functions :py:class:`haven.energy_ranges.ERange` and
    :py:class:`haven.energy_ranges.KRange`:

    .. code-block:: python

        energies = [
          ERange(13000, 13100, E_step=10),
          ERange(13100, 13160, E_step=0.5),
          KRange(13160, k_max=8, k_step=0.2),
        ]
        RE(energy_scan(energies, exposure=0.5))

    The results can also be viewed by passing
    :py:class:`haven.callbacks.live_xafs_plot.LiveXAFSPlot` to the RunEngine:

    .. code-block:: python

        RE(energy_scan(...), LiveXAFSPlot())

    Parameters
    ==========
    energies
      The X-ray energies, in eV, over which to scan.
    exposure
      How long, in seconds, to count at each energy.
    E0
      An edge energy, in eV, or name of the edge of the form
      ``"Ni_L3"``. All energies will be relative to this value.
    detectors
      The detectors to collect X-ray signal from at each energy.
    energy_signals
      Positioners that will receive the changing energies.
    time_signals
      Positioners that will receive the exposure time for each scan.
    md
      Additional metadata to pass on the to run engine.

    Yields
    ======
    Bluesky messages to execute the scan.

    """
    # Check that arguments are sensible
    if len(energy_signals) < 1:
        msg = "Cannot run energy_scan with empty *energy_signals*."
        log.error(msg)
        raise ValueError(msg)
    # Resolve the detector and positioner list if given by name
    if isinstance(detectors, str):
        detectors = beamline.registry.findall(detectors)
    real_detectors = []
    for det in detectors:
        real_detectors.extend(beamline.registry.findall(det))
    log.debug(f"Found registered detectors: {real_detectors}")
    energy_signals = [beamline.registry.find(ep) for ep in energy_signals]
    # Figure out which time positioners to use
    if time_signals is None:
        time_signals = [
            det.default_time_signal
            for det in detectors
            if hasattr(det, "default_time_signal")
        ]
    else:
        time_signals = [beamline.registry.find(tp) for tp in time_signals]
    # Convert an individual exposure time to an array of exposure times
    if not hasattr(exposure, "__iter__"):
        exposure = [exposure] * len(energies)
    # Correct for E0
    if isinstance(E0, str):
        # Look up E0 in the database if e.g. "Ni_K" is given as E0
        E0_str = E0
        E0 = edge_energy(E0)
    else:
        E0_str = None
    energies = np.asarray(energies)
    energies += E0
    # Todo: sort the energies and exposure times by the energy
    # Prepare the positioners list with associated energies and exposures
    scan_args = [(motor, energies) for motor in energy_signals]
    scan_args += [(motor, exposure) for motor in time_signals]
    scan_args = [item for items in scan_args for item in items]
    # Add some extra metadata
    config = load_config()
    md_ = {"edge": E0_str, "E0": E0}
    # Do the actual scan
    yield from bp.list_scan(
        real_detectors,
        *scan_args,
        md=ChainMap(md, md_, config),
    )


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from collections import OrderedDict
from unittest import mock

import numpy as np
import pytest
from ophyd import StatusBase

from haven import exceptions
from haven.devices.aerotech import AerotechMotor, AerotechStage, ureg


@pytest.fixture()
async def aerotech():
    stage = AerotechStage(
        horizontal_prefix="255idc:m1",
        vertical_prefix="255idc:m2",
        name="aerotech",
    )
    await stage.connect(mock=True)
    return stage


@pytest.fixture()
def aerotech_axis(aerotech):
    m = aerotech.horiz
    yield m


def test_aerotech_flyer(sim_registry):
    aeroflyer = AerotechMotor(
        prefix="255idc:m1", name="aerotech_flyer", axis="@0", encoder=6
    )
    assert aeroflyer is not None


async def test_aerotech_stage(sim_registry):
    fly_stage = AerotechStage(
        vertical_prefix="255idc:m1",
        horizontal_prefix="255idc:m2",
        name="aerotech",
    )
    assert fly_stage is not None
    # assert fly_stage.asyn.ascii_output.pvname == "motor_ioc:asynEns.AOUT"


@pytest.mark.skip(reason="Aerotech support needs to be re-written for new hardware")
def test_aerotech_fly_params_forward(aerotech_flyer):
    flyer = aerotech_flyer
    # Set some example positions
    flyer.motor_egu.put("micron")
    flyer.acceleration.put(0.5)  # sec
    flyer.encoder_resolution.put(0.001)  # µm
    flyer.flyer_start_position.put(10.0)  # µm
    flyer.flyer_end_position.put(20.0)  # µm
    flyer.flyer_num_points.put(101)  # µm
    flyer.flyer_dwell_time.put(1)  # sec

    # Check that the fly-scan parameters were calculated correctly
    assert flyer.pso_start.get(use_monitor=False) == 9.95
    assert flyer.pso_end.get(use_monitor=False) == 20.05
    assert flyer.flyer_slew_speed.get(use_monitor=False) == 0.1  # µm/sec
    assert flyer.flyer_taxi_start.get(use_monitor=False) == 9.85  # µm
    assert flyer.flyer_taxi_end.get(use_monitor=False) == pytest.approx(20.0875)  # µm
    assert flyer.encoder_step_size.get(use_monitor=False) == 100
    assert flyer.encoder_window_start.get(use_monitor=False) == -5
    assert flyer.encoder_window_end.get(use_monitor=False) == 10105
    i = 10.0
    pixel = []
    while i <= 20.03:
        pixel.append(i)
        i = i + 0.1
    np.testing.assert_allclose(flyer.pixel_positions, pixel)


@pytest.mark.skip(reason="Aerotech support needs to be re-written for new hardware")
def test_aerotech_fly_params_reverse(aerotech_flyer):
    flyer = aerotech_flyer
    # Set some example positions
    flyer.motor_egu.put("micron")
    flyer.acceleration.put(0.5)  # sec
    flyer.encoder_resolution.put(0.001)  # µm
    flyer.flyer_start_position.put(20.0)  # µm
    flyer.flyer_end_position.put(10.0)  # µm
    flyer.flyer_num_points.put(101)  # µm
    flyer.flyer_dwell_time.put(1)  # sec

    # Check that the fly-scan parameters were calculated correctly
    assert flyer.pso_start.get(use_monitor=False) == 20.05
    assert flyer.pso_end.get(use_monitor=False) == 9.95
    assert flyer.flyer_slew_speed.get(use_monitor=False) == 0.1  # µm/sec
    assert flyer.flyer_taxi_start.get(use_monitor=False) == pytest.approx(20.15)  # µm
    assert flyer.flyer_taxi_end.get(use_monitor=False) == 9.9125  # µm
    assert flyer.encoder_step_size.get(use_monitor=False) == 100
    assert flyer.encoder_window_start.get(use_monitor=False) == 5
    assert flyer.encoder_window_end.get(use_monitor=False) == -10105


@pytest.mark.skip(reason="Aerotech support needs to be re-written for new hardware")
def test_aerotech_fly_params_no_window(aerotech_flyer):
    """Test the fly scan params when the range is too large for the PSO window."""
    flyer = aerotech_flyer
    # Set some example positions
    flyer.motor_egu.put("micron")
    flyer.acceleration.put(0.5)  # sec
    flyer.encoder_resolution.put(0.001)  # µm
    flyer.flyer_start_position.put(0)  # µm
    flyer.flyer_end_position.put(9000)  # µm
    flyer.flyer_num_points.put(90001)  # µm
    flyer.flyer_dwell_time.put(1)  # sec

    # Check that the fly-scan parameters were calculated correctly
    assert flyer.pso_start.get(use_monitor=False) == -0.05
    assert flyer.pso_end.get(use_monitor=False) == 9000.05
    assert flyer.flyer_taxi_start.get(use_monitor=False) == pytest.approx(-0.15)  # µm
    assert flyer.flyer_taxi_end.get(use_monitor=False) == 9000.0875  # µm
    assert flyer.encoder_step_size.get(use_monitor=False) == 100
    assert flyer.encoder_window_start.get(use_monitor=False) == -5
    assert flyer.encoder_window_end.get(use_monitor=False) == 9000105
    assert flyer.encoder_use_window.get(use_monitor=False) is False


@pytest.mark.skip(reason="Aerotech support needs to be re-written for new hardware")
def test_aerotech_predicted_positions(aerotech_flyer):
    """Check that the fly-scan positions are calculated properly."""
    flyer = aerotech_flyer
    # Set some example positions
    flyer.motor_egu.put("micron")
    flyer.acceleration.put(0.5)  # sec
    flyer.encoder_resolution.put(0.001)  # µm
    flyer.flyer_start_position.put(10.05)  # µm
    flyer.flyer_end_position.put(19.95)  # µm
    flyer.flyer_num_points.put(100)  # µm
    flyer.flyer_dwell_time.put(1)  # sec

    # Check that the fly-scan parameters were calculated correctly
    i = 10.05
    pixel_positions = []
    while i <= 19.98:
        pixel_positions.append(i)
        i = i + 0.1
    num_pulses = len(pixel_positions) + 1
    pso_positions = np.linspace(10, 20, num=num_pulses)
    encoder_pso_positions = np.linspace(0, 10000, num=num_pulses)
    np.testing.assert_allclose(flyer.encoder_pso_positions, encoder_pso_positions)
    np.testing.assert_allclose(flyer.pso_positions, pso_positions)
    np.testing.assert_allclose(flyer.pixel_positions, pixel_positions)


@pytest.mark.skip(reason="Aerotech support needs to be re-written for new hardware")
def test_enable_pso(aerotech_flyer):
    flyer = aerotech_flyer
    # Set up scan parameters
    flyer.encoder_step_size.put(50)  # In encoder counts
    flyer.encoder_window_start.put(-5)  # In encoder counts
    flyer.encoder_window_end.put(10000)  # In encoder counts
    flyer.encoder_use_window.put(True)
    # Check that commands are sent to set up the controller for flying
    flyer.enable_pso()
    assert flyer.send_command.called
    commands = [c.args[0] for c in flyer.send_command.call_args_list]
    assert commands == [
        "PSOCONTROL @0 RESET",
        "PSOOUTPUT @0 CONTROL 1",
        "PSOPULSE @0 TIME 20, 10",
        "PSOOUTPUT @0 PULSE WINDOW MASK",
        "PSOTRACK @0 INPUT 6",
        "PSODISTANCE @0 FIXED 50",
        "PSOWINDOW @0 1 INPUT 6",
        "PSOWINDOW @0 1 RANGE -5,10000",
    ]


@pytest.mark.skip(reason="Aerotech support needs to be re-written for new hardware")
def test_enable_pso_no_window(aerotech_flyer):
    flyer = aerotech_flyer
    # Set up scan parameters
    flyer.encoder_step_size.put(50)  # In encoder counts
    flyer.encoder_window_start.put(-5)  # In encoder counts
    flyer.encoder_window_end.put(None)  # High end is outside the window range
    # Check that commands are sent to set up the controller for flying
    flyer.enable_pso()
    assert flyer.send_command.called
    commands = [c.args[0] for c in flyer.send_command.call_args_list]
    assert commands == [
        "PSOCONTROL @0 RESET",
        "PSOOUTPUT @0 CONTROL 1",
        "PSOPULSE @0 TIME 20, 10",
        "PSOOUTPUT @0 PULSE",
        "PSOTRACK @0 INPUT 6",
        "PSODISTANCE @0 FIXED 50",
        # "PSOWINDOW @0 1 INPUT 6",
        # "PSOWINDOW @0 1 RANGE -5,10000",
    ]


@pytest.mark.skip(reason="Aerotech support needs to be re-written for new hardware")
def test_pso_bad_window_forward(aerotech_flyer):
    """Check for an exception when the window is needed but not enabled.

    I.e. when the taxi distance is larger than the encoder step size."""
    flyer = aerotech_flyer
    # Set up scan parameters
    flyer.encoder_resolution.put(1)
    flyer.encoder_step_size.put(5 / flyer.encoder_resolution.get())  # In encoder counts
    flyer.encoder_window_start.put(-5)  # In encoder counts
    flyer.encoder_window_end.put(None)  # High end is outside the window range
    flyer.pso_end.put(100)
    flyer.flyer_taxi_end.put(110)
    # Check that commands are sent to set up the controller for flying
    with pytest.raises(exceptions.InvalidScanParameters):
        flyer.enable_pso()


@pytest.mark.skip(reason="Aerotech support needs to be re-written for new hardware")
def test_pso_bad_window_reverse(aerotech_flyer):
    """Check for an exception when the window is needed but not enabled.

    I.e. when the taxi distance is larger than the encoder step size."""
    flyer = aerotech_flyer
    # Set up scan parameters
    flyer.encoder_resolution.put(1)
    flyer.flyer_end_position.put(5)
    flyer.flyer_num_points.put(2)
    flyer.encoder_step_size.put(
        flyer.flyer_step_size() / flyer.encoder_resolution.get()
    )  # In encoder counts
    flyer.encoder_window_start.put(114)  # In encoder counts
    flyer.encoder_window_start.put(None)  # High end is outside the window range
    flyer.pso_start.put(100)
    flyer.flyer_taxi_start.put(94)
    # Check that commands are sent to set up the controller for flying
    with pytest.raises(exceptions.InvalidScanParameters):
        flyer.enable_pso()


@pytest.mark.skip(reason="Aerotech support needs to be re-written for new hardware")
def test_arm_pso(aerotech_flyer):
    flyer = aerotech_flyer
    assert not flyer.send_command.called
    flyer.arm_pso()
    assert flyer.send_command.called
    command = flyer.send_command.call_args.args[0]
    assert command == "PSOCONTROL @0 ARM"


@pytest.mark.skip(reason="Aerotech support needs to be re-written for new hardware")
def test_motor_units(aerotech_flyer):
    """Check that the motor and flyer handle enginering units properly."""
    flyer = aerotech_flyer
    flyer.motor_egu.put("micron")
    unit = flyer.motor_egu_pint
    assert unit == ureg("1e-6 m")


@pytest.mark.skip(reason="Aerotech support needs to be re-written for new hardware")
def test_kickoff(aerotech_flyer):
    # Set up fake flyer with mocked fly method
    flyer = aerotech_flyer
    flyer.taxi = mock.MagicMock()
    flyer.flyer_dwell_time.put(1.0)
    # Start flying
    status = flyer.kickoff()
    # Check status behavior matches flyer interface
    assert isinstance(status, StatusBase)
    assert not status.done
    # Start flying and see if the status is done
    flyer.ready_to_fly.put(True)
    status.wait()
    assert status.done
    assert type(flyer.starttime) == float


@pytest.mark.skip(reason="Aerotech support needs to be re-written for new hardware")
def test_complete(aerotech_flyer):
    # Set up fake flyer with mocked fly method
    flyer = aerotech_flyer
    flyer.move = mock.MagicMock()
    assert flyer.user_setpoint.get() == 0
    flyer.flyer_taxi_end.put(10)
    # Complete flying
    status = flyer.complete()
    # Check that the motor was moved
    assert flyer.move.called_with(9)
    # Check status behavior matches flyer interface
    assert isinstance(status, StatusBase)
    status.wait(timeout=1)
    assert status.done


@pytest.mark.skip(reason="Aerotech support needs to be re-written for new hardware")
def test_collect(aerotech_flyer):
    flyer = aerotech_flyer
    # Set up needed parameters
    flyer.pixel_positions = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    flyer.starttime = 0
    flyer.endtime = flyer.starttime + 11.25
    flyer.acceleration.put(0.5)  # µm/s^2
    flyer.flyer_end_position.put(0.1)
    flyer.flyer_num_points.put(2)  # µm
    flyer.flyer_dwell_time.put(1)  # sec
    expected_timestamps = [
        1.125,
        2.125,
        3.125,
        4.125,
        5.125,
        6.125,
        7.125,
        8.125,
        9.125,
        10.125,
    ]
    payload = list(flyer.collect())
    # Confirm data have the right structure
    for datum, value, timestamp in zip(
        payload, flyer.pixel_positions, expected_timestamps
    ):
        assert datum == {
            "data": {
                "aerotech_horiz": value,
                "aerotech_horiz_user_setpoint": value,
            },
            "timestamps": {
                "aerotech_horiz": timestamp,
                "aerotech_horiz_user_setpoint": timestamp,
            },
            "time": timestamp,
        }


@pytest.mark.skip(reason="Aerotech support needs to be re-written for new hardware")
def test_describe_collect(aerotech_flyer):
    expected = {
        "positions": OrderedDict(
            [
                (
                    "aerotech_horiz",
                    {
                        "source": "SIM:aerotech_horiz",
                        "dtype": "integer",
                        "shape": [],
                        "precision": 3,
                    },
                ),
                (
                    "aerotech_horiz_user_setpoint",
                    {
                        "source": "SIM:aerotech_horiz_user_setpoint",
                        "dtype": "integer",
                        "shape": [],
                        "precision": 3,
                    },
                ),
            ]
        )
    }

    assert aerotech_flyer.describe_collect() == expected


@pytest.mark.skip(reason="Aerotech support needs to be re-written for new hardware")
def test_fly_motor_positions(aerotech_flyer):
    flyer = aerotech_flyer
    # Arbitrary rest position
    flyer.user_setpoint.put(255)
    flyer.parent.delay.channel_C.delay.sim_put(1.5)
    flyer.parent.delay.output_CD.polarity.sim_put(1)
    # Set example fly scan parameters
    flyer.flyer_taxi_start.put(5)
    flyer.flyer_start_position.put(10)
    flyer.pso_start.put(9.5)
    flyer.flyer_taxi_end.put(105)
    flyer.encoder_use_window.put(True)
    # Mock the motor position so that it returns a status we control
    motor_status = StatusBase()
    motor_status.set_finished()
    mover = mock.MagicMock(return_value=motor_status)
    flyer.move = mover
    # Check the fly scan moved the motors in the right order
    flyer.taxi()
    flyer.fly()
    assert mover.called
    positions = [c.args[0] for c in mover.call_args_list]
    assert len(positions) == 3
    pso_arm, taxi, end = positions
    assert pso_arm == 9.5
    assert taxi == 5
    assert end == 105
    # Check that the delay generator is properly configured
    assert flyer.parent.delay.channel_C.delay.get(use_monitor=False) == 0.0
    assert flyer.parent.delay.output_CD.polarity.get(use_monitor=False) == 0


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging

from bluesky_queueserver_api import BPlan
from qasync import asyncSlot
from qtpy import QtWidgets
from qtpy.QtGui import QDoubleValidator

from firefly.component_selector import ComponentSelector
from firefly.plans import regions_display

log = logging.getLogger()


class LineScanRegion(regions_display.RegionBase):

    def setup_ui(self):
        self.layout = QtWidgets.QHBoxLayout()

        # First item, ComponentSelector
        self.motor_box = ComponentSelector()
        self.layout.addWidget(self.motor_box)

        # Second item, start point
        self.start_line_edit = QtWidgets.QLineEdit()
        self.start_line_edit.setValidator(QDoubleValidator())  # only takes floats
        self.start_line_edit.setPlaceholderText("Start…")
        self.layout.addWidget(self.start_line_edit)

        # Third item, stop point
        self.stop_line_edit = QtWidgets.QLineEdit()
        self.stop_line_edit.setValidator(QDoubleValidator())  # only takes floats
        self.stop_line_edit.setPlaceholderText("Stop…")
        self.layout.addWidget(self.stop_line_edit)


class LineScanDisplay(regions_display.RegionsDisplay):
    Region = LineScanRegion

    @asyncSlot(object)
    async def update_devices_slot(self, registry):
        await self.update_devices(registry)
        await self.detectors_list.update_devices(registry)

    def time_per_scan(self, detector_time):
        num_points = self.ui.scan_pts_spin_box.value()
        total_time_per_scan = detector_time * num_points
        return total_time_per_scan

    def customize_ui(self):
        super().customize_ui()
        # When selections of detectors changed update_total_time
        self.ui.scan_pts_spin_box.valueChanged.connect(self.update_total_time)
        self.ui.detectors_list.selectionModel().selectionChanged.connect(
            self.update_total_time
        )
        self.ui.spinBox_repeat_scan_num.valueChanged.connect(self.update_total_time)

    def queue_plan(self, *args, **kwargs):
        """Execute this plan on the queueserver."""
        detectors, motor_args, repeat_scan_num = self.get_scan_parameters()
        md = self.get_meta_data()
        num_points = self.ui.scan_pts_spin_box.value()

        if self.ui.relative_scan_checkbox.isChecked():
            if self.ui.log_scan_checkbox.isChecked():
                scan_type = "rel_log_scan"
            else:
                scan_type = "rel_scan"
        else:
            if self.ui.log_scan_checkbox.isChecked():
                scan_type = "log_scan"
            else:
                scan_type = "scan"

        # # Build the queue item
        item = BPlan(
            scan_type,
            detectors,
            *motor_args,
            num=num_points,
            # per_step=None,
            md=md,
        )

        # Submit the item to the queueserver
        log.info("Added line scan() plan to queue.")
        # repeat scans
        for i in range(repeat_scan_num):
            self.queue_item_submitted.emit(item)

    def ui_filename(self):
        return "plans/line_scan.ui"


# -----------------------------------------------------------------------------
# :author:    Juanjuan Huang
# :email:     juanjuan.huang@anl.gov
# :copyright: Copyright © 2024, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------
